from typing import ClassVar, Dict, Optional, List

from attrs import define, field

from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec, GraphBuilder
from resoto_plugin_aws.resource.kms import AwsKmsKey
from resotolib.baseresources import ModelReference
from resotolib.json_bender import Bender, S, Bend, ForallBend, K, bend
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.utils import ToDict
from typing import Type
from datetime import datetime
from resotolib.types import Json
from resoto_plugin_aws.resource.ec2 import AwsEc2SecurityGroup


# noinspection PyUnresolvedReferences
class ElastiCacheTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service="elasticache",
                action="add-tags-to-resource",
                result_name=None,
                ResourceName=self.arn,
                Tags=[{"Key": key, "Value": value}],
            )
            return True
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service="elasticache",
                action="remove-tags-from-resource",
                result_name=None,
                ResourceName=self.arn,
                TagKeys=[key],
            )
            return True
        return False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("elasticache", "add-tags-to-resource"),
            AwsApiSpec("elasticache", "remove-tags-from-resource"),
        ]


@define(eq=False, slots=False)
class AwsElastiCacheEndpoint:
    kind: ClassVar[str] = "aws_elasticache_endpoint"
    mapping: ClassVar[Dict[str, Bender]] = {"address": S("Address"), "port": S("Port")}
    address: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheDestinationDetails:
    kind: ClassVar[str] = "aws_elasticache_destination_details"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cloud_watch_logs_details": S("CloudWatchLogsDetails", "LogGroup"),
        "kinesis_firehose_details": S("KinesisFirehoseDetails", "DeliveryStream"),
    }
    cloud_watch_logs_details: Optional[str] = field(default=None)
    kinesis_firehose_details: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCachePendingLogDeliveryConfiguration:
    kind: ClassVar[str] = "aws_elasticache_pending_log_delivery_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "log_type": S("LogType"),
        "destination_type": S("DestinationType"),
        "destination_details": S("DestinationDetails") >> Bend(AwsElastiCacheDestinationDetails.mapping),
        "log_format": S("LogFormat"),
    }
    log_type: Optional[str] = field(default=None)
    destination_type: Optional[str] = field(default=None)
    destination_details: Optional[AwsElastiCacheDestinationDetails] = field(default=None)
    log_format: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCachePendingModifiedValues:
    kind: ClassVar[str] = "aws_elasticache_pending_modified_values"
    mapping: ClassVar[Dict[str, Bender]] = {
        "num_cache_nodes": S("NumCacheNodes"),
        "cache_node_ids_to_remove": S("CacheNodeIdsToRemove", default=[]),
        "engine_version": S("EngineVersion"),
        "cache_node_type": S("CacheNodeType"),
        "auth_token_status": S("AuthTokenStatus"),
        "log_delivery_configurations": S("LogDeliveryConfigurations", default=[])
        >> ForallBend(AwsElastiCachePendingLogDeliveryConfiguration.mapping),
    }
    num_cache_nodes: Optional[int] = field(default=None)
    cache_node_ids_to_remove: List[str] = field(factory=list)
    engine_version: Optional[str] = field(default=None)
    cache_node_type: Optional[str] = field(default=None)
    auth_token_status: Optional[str] = field(default=None)
    log_delivery_configurations: List[AwsElastiCachePendingLogDeliveryConfiguration] = field(factory=list)


@define(eq=False, slots=False)
class AwsElastiCacheNotificationConfiguration:
    kind: ClassVar[str] = "aws_elasticache_notification_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {"topic_arn": S("TopicArn"), "topic_status": S("TopicStatus")}
    topic_arn: Optional[str] = field(default=None)
    topic_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheCacheSecurityGroupMembership:
    kind: ClassVar[str] = "aws_elasticache_cache_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cache_security_group_name": S("CacheSecurityGroupName"),
        "status": S("Status"),
    }
    cache_security_group_name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheCacheParameterGroupStatus:
    kind: ClassVar[str] = "aws_elasticache_cache_parameter_group_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cache_parameter_group_name": S("CacheParameterGroupName"),
        "parameter_apply_status": S("ParameterApplyStatus"),
        "cache_node_ids_to_reboot": S("CacheNodeIdsToReboot", default=[]),
    }
    cache_parameter_group_name: Optional[str] = field(default=None)
    parameter_apply_status: Optional[str] = field(default=None)
    cache_node_ids_to_reboot: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsElastiCacheCacheNode:
    kind: ClassVar[str] = "aws_elasticache_cache_node"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cache_node_id": S("CacheNodeId"),
        "cache_node_status": S("CacheNodeStatus"),
        "cache_node_create_time": S("CacheNodeCreateTime"),
        "endpoint": S("Endpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "parameter_group_status": S("ParameterGroupStatus"),
        "source_cache_node_id": S("SourceCacheNodeId"),
        "customer_availability_zone": S("CustomerAvailabilityZone"),
        "customer_outpost_arn": S("CustomerOutpostArn"),
    }
    cache_node_id: Optional[str] = field(default=None)
    cache_node_status: Optional[str] = field(default=None)
    cache_node_create_time: Optional[datetime] = field(default=None)
    endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    parameter_group_status: Optional[str] = field(default=None)
    source_cache_node_id: Optional[str] = field(default=None)
    customer_availability_zone: Optional[str] = field(default=None)
    customer_outpost_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheSecurityGroupMembership:
    kind: ClassVar[str] = "aws_elasticache_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {"security_group_id": S("SecurityGroupId"), "status": S("Status")}
    security_group_id: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheLogDeliveryConfiguration:
    kind: ClassVar[str] = "aws_elasticache_log_delivery_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "log_type": S("LogType"),
        "destination_type": S("DestinationType"),
        "destination_details": S("DestinationDetails") >> Bend(AwsElastiCacheDestinationDetails.mapping),
        "log_format": S("LogFormat"),
        "status": S("Status"),
        "message": S("Message"),
    }
    log_type: Optional[str] = field(default=None)
    destination_type: Optional[str] = field(default=None)
    destination_details: Optional[AwsElastiCacheDestinationDetails] = field(default=None)
    log_format: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheCacheCluster(ElastiCacheTaggable, AwsResource):
    kind: ClassVar[str] = "aws_elasticache_cache_cluster"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elasticache", "describe-cache-clusters", "CacheClusters")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("CacheClusterId"),
        "tags": S("Tags", default=[]) >> ToDict(),
        "name": S("CacheClusterId"),
        "arn": S("ARN"),
        "ctime": S("CacheClusterCreateTime"),
        "mtime": K(None),
        "atime": K(None),
        "cluster_configuration_endpoint": S("ConfigurationEndpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "cluster_client_download_landing_page": S("ClientDownloadLandingPage"),
        "cluster_cache_node_type": S("CacheNodeType"),
        "cluster_engine": S("Engine"),
        "cluster_engine_version": S("EngineVersion"),
        "cluster_cache_cluster_status": S("CacheClusterStatus"),
        "cluster_num_cache_nodes": S("NumCacheNodes"),
        "cluster_preferred_availability_zone": S("PreferredAvailabilityZone"),
        "cluster_preferred_outpost_arn": S("PreferredOutpostArn"),
        "cluster_cache_cluster_create_time": S("CacheClusterCreateTime"),
        "cluster_preferred_maintenance_window": S("PreferredMaintenanceWindow"),
        "cluster_pending_modified_values": S("PendingModifiedValues")
        >> Bend(AwsElastiCachePendingModifiedValues.mapping),
        "cluster_notification_configuration": S("NotificationConfiguration")
        >> Bend(AwsElastiCacheNotificationConfiguration.mapping),
        "cluster_cache_security_groups": S("CacheSecurityGroups", default=[])
        >> ForallBend(AwsElastiCacheCacheSecurityGroupMembership.mapping),
        "cluster_cache_parameter_group": S("CacheParameterGroup")
        >> Bend(AwsElastiCacheCacheParameterGroupStatus.mapping),
        "cluster_cache_subnet_group_name": S("CacheSubnetGroupName"),
        "cluster_cache_nodes": S("CacheNodes", default=[]) >> ForallBend(AwsElastiCacheCacheNode.mapping),
        "cluster_auto_minor_version_upgrade": S("AutoMinorVersionUpgrade"),
        "cluster_security_groups": S("SecurityGroups", default=[])
        >> ForallBend(AwsElastiCacheSecurityGroupMembership.mapping),
        "cluster_replication_group_id": S("ReplicationGroupId"),
        "cluster_snapshot_retention_limit": S("SnapshotRetentionLimit"),
        "cluster_snapshot_window": S("SnapshotWindow"),
        "cluster_auth_token_enabled": S("AuthTokenEnabled"),
        "cluster_auth_token_last_modified_date": S("AuthTokenLastModifiedDate"),
        "cluster_transit_encryption_enabled": S("TransitEncryptionEnabled"),
        "cluster_at_rest_encryption_enabled": S("AtRestEncryptionEnabled"),
        "cluster_replication_group_log_delivery_enabled": S("ReplicationGroupLogDeliveryEnabled"),
        "cluster_log_delivery_configurations": S("LogDeliveryConfigurations", default=[])
        >> ForallBend(AwsElastiCacheLogDeliveryConfiguration.mapping),
    }
    cluster_configuration_endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    cluster_client_download_landing_page: Optional[str] = field(default=None)
    cluster_cache_node_type: Optional[str] = field(default=None)
    cluster_engine: Optional[str] = field(default=None)
    cluster_engine_version: Optional[str] = field(default=None)
    cluster_cache_cluster_status: Optional[str] = field(default=None)
    cluster_num_cache_nodes: Optional[int] = field(default=None)
    cluster_preferred_availability_zone: Optional[str] = field(default=None)
    cluster_preferred_outpost_arn: Optional[str] = field(default=None)
    cluster_preferred_maintenance_window: Optional[str] = field(default=None)
    cluster_pending_modified_values: Optional[AwsElastiCachePendingModifiedValues] = field(default=None)
    cluster_notification_configuration: Optional[AwsElastiCacheNotificationConfiguration] = field(default=None)
    cluster_cache_security_groups: List[AwsElastiCacheCacheSecurityGroupMembership] = field(factory=list)
    cluster_cache_parameter_group: Optional[AwsElastiCacheCacheParameterGroupStatus] = field(default=None)
    cluster_cache_subnet_group_name: Optional[str] = field(default=None)
    cluster_cache_nodes: List[AwsElastiCacheCacheNode] = field(factory=list)
    cluster_auto_minor_version_upgrade: Optional[bool] = field(default=None)
    cluster_security_groups: List[AwsElastiCacheSecurityGroupMembership] = field(factory=list)
    cluster_replication_group_id: Optional[str] = field(default=None)
    cluster_snapshot_retention_limit: Optional[int] = field(default=None)
    cluster_snapshot_window: Optional[str] = field(default=None)
    cluster_auth_token_enabled: Optional[bool] = field(default=None)
    cluster_auth_token_last_modified_date: Optional[datetime] = field(default=None)
    cluster_transit_encryption_enabled: Optional[bool] = field(default=None)
    cluster_at_rest_encryption_enabled: Optional[bool] = field(default=None)
    cluster_replication_group_log_delivery_enabled: Optional[bool] = field(default=None)
    cluster_log_delivery_configurations: List[AwsElastiCacheLogDeliveryConfiguration] = field(factory=list)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            aws_service=self.api_spec.service, action="delete-cache-cluster", result_name=None, CacheClusterId=self.id
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [AwsApiSpec("elasticache", "delete-cache-cluster")]

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [cls.api_spec, AwsApiSpec(cls.api_spec.service, "list-tags-for-resource")]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(resource: AwsElastiCacheCacheCluster) -> None:
            tags = builder.client.list(
                resource.api_spec.service, "list-tags-for-resource", "TagList", ResourceName=resource.arn
            )
            if tags:
                resource.tags = bend(ToDict(), tags)

        for js in json:
            instance = cls.from_api(js)
            builder.add_node(instance, js)
            builder.submit_work(add_tags, instance)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        for sg in self.cluster_security_groups:
            builder.dependant_node(self, reverse=True, clazz=AwsEc2SecurityGroup, id=sg.security_group_id)


@define(eq=False, slots=False)
class AwsElastiCacheGlobalReplicationGroupInfo:
    kind: ClassVar[str] = "aws_elasticache_global_replication_group_info"
    mapping: ClassVar[Dict[str, Bender]] = {
        "global_replication_group_id": S("GlobalReplicationGroupId"),
        "global_replication_group_member_role": S("GlobalReplicationGroupMemberRole"),
    }
    global_replication_group_id: Optional[str] = field(default=None)
    global_replication_group_member_role: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheReshardingStatus:
    kind: ClassVar[str] = "aws_elasticache_resharding_status"
    mapping: ClassVar[Dict[str, Bender]] = {"slot_migration": S("SlotMigration", "ProgressPercentage")}
    slot_migration: Optional[float] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheUserGroupsUpdateStatus:
    kind: ClassVar[str] = "aws_elasticache_user_groups_update_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "user_group_ids_to_add": S("UserGroupIdsToAdd", default=[]),
        "user_group_ids_to_remove": S("UserGroupIdsToRemove", default=[]),
    }
    user_group_ids_to_add: List[str] = field(factory=list)
    user_group_ids_to_remove: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsElastiCacheReplicationGroupPendingModifiedValues:
    kind: ClassVar[str] = "aws_elasticache_replication_group_pending_modified_values"
    mapping: ClassVar[Dict[str, Bender]] = {
        "primary_cluster_id": S("PrimaryClusterId"),
        "automatic_failover_status": S("AutomaticFailoverStatus"),
        "resharding": S("Resharding") >> Bend(AwsElastiCacheReshardingStatus.mapping),
        "auth_token_status": S("AuthTokenStatus"),
        "user_groups": S("UserGroups") >> Bend(AwsElastiCacheUserGroupsUpdateStatus.mapping),
        "log_delivery_configurations": S("LogDeliveryConfigurations", default=[])
        >> ForallBend(AwsElastiCachePendingLogDeliveryConfiguration.mapping),
    }
    primary_cluster_id: Optional[str] = field(default=None)
    automatic_failover_status: Optional[str] = field(default=None)
    resharding: Optional[AwsElastiCacheReshardingStatus] = field(default=None)
    auth_token_status: Optional[str] = field(default=None)
    user_groups: Optional[AwsElastiCacheUserGroupsUpdateStatus] = field(default=None)
    log_delivery_configurations: List[AwsElastiCachePendingLogDeliveryConfiguration] = field(factory=list)


@define(eq=False, slots=False)
class AwsElastiCacheNodeGroupMember:
    kind: ClassVar[str] = "aws_elasticache_node_group_member"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cache_cluster_id": S("CacheClusterId"),
        "cache_node_id": S("CacheNodeId"),
        "read_endpoint": S("ReadEndpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "preferred_availability_zone": S("PreferredAvailabilityZone"),
        "preferred_outpost_arn": S("PreferredOutpostArn"),
        "current_role": S("CurrentRole"),
    }
    cache_cluster_id: Optional[str] = field(default=None)
    cache_node_id: Optional[str] = field(default=None)
    read_endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    preferred_availability_zone: Optional[str] = field(default=None)
    preferred_outpost_arn: Optional[str] = field(default=None)
    current_role: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElastiCacheNodeGroup:
    kind: ClassVar[str] = "aws_elasticache_node_group"
    mapping: ClassVar[Dict[str, Bender]] = {
        "node_group_id": S("NodeGroupId"),
        "status": S("Status"),
        "primary_endpoint": S("PrimaryEndpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "reader_endpoint": S("ReaderEndpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "slots": S("Slots"),
        "node_group_members": S("NodeGroupMembers", default=[]) >> ForallBend(AwsElastiCacheNodeGroupMember.mapping),
    }
    node_group_id: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)
    primary_endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    reader_endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    slots: Optional[str] = field(default=None)
    node_group_members: List[AwsElastiCacheNodeGroupMember] = field(factory=list)


@define(eq=False, slots=False)
class AwsElastiCacheReplicationGroup(ElastiCacheTaggable, AwsResource):
    kind: ClassVar[str] = "aws_elasticache_replication_group"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elasticache", "describe-replication-groups", "ReplicationGroups")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"delete": ["aws_elasticache_cache_cluster", "aws_kms_key"]},
        "successors": {"default": ["aws_elasticache_cache_cluster", "aws_kms_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("ReplicationGroupId"),
        "name": S("ReplicationGroupId"),
        "arn": S("ARN"),
        "ctime": S("ReplicationGroupCreateTime"),
        "mtime": K(None),
        "atime": K(None),
        "replication_group_description": S("Description"),
        "replication_group_global_replication_group_info": S("GlobalReplicationGroupInfo")
        >> Bend(AwsElastiCacheGlobalReplicationGroupInfo.mapping),
        "replication_group_status": S("Status"),
        "replication_group_pending_modified_values": S("PendingModifiedValues")
        >> Bend(AwsElastiCacheReplicationGroupPendingModifiedValues.mapping),
        "replication_group_member_clusters": S("MemberClusters", default=[]),
        "replication_group_node_groups": S("NodeGroups", default=[]) >> ForallBend(AwsElastiCacheNodeGroup.mapping),
        "replication_group_snapshotting_cluster_id": S("SnapshottingClusterId"),
        "replication_group_automatic_failover": S("AutomaticFailover"),
        "replication_group_multi_az": S("MultiAZ"),
        "replication_group_configuration_endpoint": S("ConfigurationEndpoint") >> Bend(AwsElastiCacheEndpoint.mapping),
        "replication_group_snapshot_retention_limit": S("SnapshotRetentionLimit"),
        "replication_group_snapshot_window": S("SnapshotWindow"),
        "replication_group_cluster_enabled": S("ClusterEnabled"),
        "replication_group_cache_node_type": S("CacheNodeType"),
        "replication_group_auth_token_enabled": S("AuthTokenEnabled"),
        "replication_group_auth_token_last_modified_date": S("AuthTokenLastModifiedDate"),
        "replication_group_transit_encryption_enabled": S("TransitEncryptionEnabled"),
        "replication_group_at_rest_encryption_enabled": S("AtRestEncryptionEnabled"),
        "replication_group_member_clusters_outpost_arns": S("MemberClustersOutpostArns", default=[]),
        "replication_group_kms_key_id": S("KmsKeyId"),
        "replication_group_arn": S("ARN"),
        "replication_group_user_group_ids": S("UserGroupIds", default=[]),
        "replication_group_log_delivery_configurations": S("LogDeliveryConfigurations", default=[])
        >> ForallBend(AwsElastiCacheLogDeliveryConfiguration.mapping),
        "replication_group_data_tiering": S("DataTiering"),
    }
    replication_group_description: Optional[str] = field(default=None)
    replication_group_global_replication_group_info: Optional[AwsElastiCacheGlobalReplicationGroupInfo] = field(
        default=None
    )
    replication_group_status: Optional[str] = field(default=None)
    replication_group_pending_modified_values: Optional[AwsElastiCacheReplicationGroupPendingModifiedValues] = field(
        default=None
    )
    replication_group_member_clusters: List[str] = field(factory=list)
    replication_group_node_groups: List[AwsElastiCacheNodeGroup] = field(factory=list)
    replication_group_snapshotting_cluster_id: Optional[str] = field(default=None)
    replication_group_automatic_failover: Optional[str] = field(default=None)
    replication_group_multi_az: Optional[str] = field(default=None)
    replication_group_configuration_endpoint: Optional[AwsElastiCacheEndpoint] = field(default=None)
    replication_group_snapshot_retention_limit: Optional[int] = field(default=None)
    replication_group_snapshot_window: Optional[str] = field(default=None)
    replication_group_cluster_enabled: Optional[bool] = field(default=None)
    replication_group_cache_node_type: Optional[str] = field(default=None)
    replication_group_auth_token_enabled: Optional[bool] = field(default=None)
    replication_group_auth_token_last_modified_date: Optional[datetime] = field(default=None)
    replication_group_transit_encryption_enabled: Optional[bool] = field(default=None)
    replication_group_at_rest_encryption_enabled: Optional[bool] = field(default=None)
    replication_group_member_clusters_outpost_arns: List[str] = field(factory=list)
    replication_group_kms_key_id: Optional[str] = field(default=None)
    replication_group_arn: Optional[str] = field(default=None)
    replication_group_user_group_ids: List[str] = field(factory=list)
    replication_group_log_delivery_configurations: List[AwsElastiCacheLogDeliveryConfiguration] = field(factory=list)
    replication_group_data_tiering: Optional[str] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [cls.api_spec, AwsApiSpec(cls.api_spec.service, "list-tags-for-resource")]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(resource: AwsElastiCacheReplicationGroup) -> None:
            tags = builder.client.list(
                resource.api_spec.service, "list-tags-for-resource", "TagList", ResourceName=resource.arn
            )
            if tags:
                resource.tags = bend(ToDict(), tags)

        for js in json:
            instance = cls.from_api(js)
            builder.add_node(instance, js)
            builder.submit_work(add_tags, instance)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        for cluster_name in self.replication_group_member_clusters:
            builder.dependant_node(
                self,
                clazz=AwsElastiCacheCacheCluster,
                name=cluster_name,
            )
        if self.replication_group_kms_key_id:
            builder.dependant_node(self, clazz=AwsKmsKey, id=self.replication_group_kms_key_id)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="delete-replication-group",
            result_name=None,
            ReplicationGroupId=self.id,
            RetainPrimaryCluster=False,
        )
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [AwsApiSpec("elasticache", "delete-replication-group")]


resources: List[Type[AwsResource]] = [AwsElastiCacheReplicationGroup, AwsElastiCacheCacheCluster]



from collections import defaultdict
from types import SimpleNamespace
from typing import Any, Dict, cast
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.sns import AwsSnsTopic, AwsSnsSubscription, AwsSnsPlatformApplication
from test.resources import round_trip_for


def test_topics() -> None:
    first, builder = round_trip_for(AwsSnsTopic)
    assert len(builder.resources_of(AwsSnsTopic)) == 1
    assert len(first.tags) == 1


def test_tagging_topics() -> None:
    topic, _ = round_trip_for(AwsSnsTopic)

    def validate_update_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "tag-resource":
            assert kwargs["ResourceArn"] == topic.arn
            assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "untag-resource":
            assert kwargs["ResourceArn"] == topic.arn
            assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    topic.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    topic.delete_resource_tag(client, "foo")


def test_delete_topics() -> None:
    topic, _ = round_trip_for(AwsSnsTopic)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-topic"
        assert kwargs["TopicArn"] == topic.arn

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    topic.delete_resource(client)


def test_subs() -> None:
    first, builder = round_trip_for(AwsSnsSubscription)
    assert len(builder.resources_of(AwsSnsSubscription)) == 1
    assert first.subscription_confirmation_was_authenticated is False
    assert first.subscription_raw_message_delivery is True


def test_delete_subs() -> None:
    sub, _ = round_trip_for(AwsSnsSubscription)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "unsubscribe"
        assert kwargs["SubscriptionArn"] == sub.arn

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    sub.delete_resource(client)


def test_apps() -> None:
    first, builder = round_trip_for(AwsSnsPlatformApplication)
    assert len(builder.resources_of(AwsSnsPlatformApplication)) == 1
    type_count: Dict[str, int] = defaultdict(int)
    for node in builder.graph.nodes:
        type_count[node.kind] += 1
    assert type_count["aws_sns_endpoint"] == 1


def test_delete_apps() -> None:
    app, _ = round_trip_for(AwsSnsPlatformApplication)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-platform-application"
        assert kwargs["PlatformApplicationArn"] == app.arn

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    app.delete_resource(client)



from resoto_plugin_aws.resource.athena import AwsAthenaDataCatalog, AwsAthenaWorkGroup
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from resoto_plugin_aws.aws_client import AwsClient


def test_data_catalogs() -> None:
    res, builder = round_trip_for(AwsAthenaDataCatalog)
    assert len(builder.resources_of(AwsAthenaDataCatalog)) == 1
    assert len(res.tags) == 1
    assert res.arn == "arn:aws:athena:eu-central-1:test:datacatalog/resoto-catalog"


def test_workgroups() -> None:
    res, builder = round_trip_for(AwsAthenaWorkGroup)
    assert len(builder.resources_of(AwsAthenaWorkGroup)) == 1
    assert len(res.tags) == 1
    assert res.arn == "arn:aws:athena:eu-central-1:test:workgroup/resoto-workgroup"


def test_data_catalogs_tagging() -> None:
    res, builder = round_trip_for(AwsAthenaDataCatalog)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag-resource"
        assert kwargs["ResourceARN"] == res.arn
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag-resource"
        assert kwargs["ResourceARN"] == res.arn
        assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    res.delete_resource_tag(client, "foo")


def test_workgroup_tagging() -> None:
    res, builder = round_trip_for(AwsAthenaWorkGroup)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag-resource"
        assert kwargs["ResourceARN"] == res.arn
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag-resource"
        assert kwargs["ResourceARN"] == res.arn
        assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    res.delete_resource_tag(client, "foo")



from typing import ClassVar, Dict, List, Type, Optional, cast, Any

from attrs import define

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec, GraphBuilder
from resoto_plugin_aws.utils import tags_as_dict
from resotolib.baseresources import BaseBucket
from resotolib.json_bender import Bender, S
from resotolib.types import Json


@define(eq=False, slots=False)
class AwsS3Bucket(AwsResource, BaseBucket):
    kind: ClassVar[str] = "aws_s3_bucket"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "s3", "list-buckets", "Buckets", override_iam_permission="s3:ListAllMyBuckets"
    )
    mapping: ClassVar[Dict[str, Bender]] = {"id": S("Name"), "name": S("Name"), "ctime": S("CreationDate")}

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [cls.api_spec, AwsApiSpec("s3", "get-bucket-tagging")]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(bucket: AwsS3Bucket, client: AwsClient) -> None:
            tags = bucket._get_tags(client)
            if tags:
                bucket.tags = cast(Dict[str, Optional[str]], tags)

        for js in json:
            bucket = cls.from_api(js)
            bucket.set_arn(builder=builder, region="", account="", resource=bucket.safe_name)
            builder.add_node(bucket, js)
            builder.submit_work(add_tags, bucket, builder.client)

    def _set_tags(self, client: AwsClient, tags: Dict[str, str]) -> bool:
        tag_set = [{"Key": k, "Value": v} for k, v in tags.items()]
        client.call(
            aws_service="s3",
            action="put-bucket-tagging",
            result_name=None,
            Bucket=self.name,
            Tagging={"TagSet": tag_set},
        )
        return True

    def _get_tags(self, client: AwsClient) -> Dict[str, str]:
        """Fetch the S3 buckets tags from the AWS API."""
        tag_list = client.list(
            aws_service="s3",
            action="get-bucket-tagging",
            result_name="TagSet",
            expected_errors=["NoSuchTagSet"],
            Bucket=self.name,
        )
        return tags_as_dict(tag_list)  # type: ignore

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        tags = self._get_tags(client)
        tags[key] = value
        return self._set_tags(client, tags)

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        tags = self._get_tags(client)
        if key in tags:
            del tags[key]
        else:
            raise KeyError(key)
        return self._set_tags(client, tags)

    def delete_resource(self, client: AwsClient) -> bool:
        def delete_bucket_content(s3: Any) -> bool:
            bucket = s3.Bucket(self.name)
            bucket.objects.delete()
            bucket.delete()
            return True

        return client.with_resource("s3", delete_bucket_content) or False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("s3", "put-bucket-tagging"),
            AwsApiSpec("s3", "delete-object"),
            AwsApiSpec("s3", "delete-bucket"),
        ]


resources: List[Type[AwsResource]] = [AwsS3Bucket]



from concurrent.futures import ThreadPoolExecutor
from typing import Sequence, Tuple, List, Optional

from resoto_plugin_aws.resource.base import ExecutorQueue


def check_executor_queue(work: Sequence[int], fail_on_first_exception: bool) -> Tuple[List[int], Optional[Exception]]:
    work_done = []

    def do_work(num: int) -> None:
        if num == 5:
            raise Exception(f"Abort {num}")
        work_done.append(num)

    with ThreadPoolExecutor(max_workers=1) as executor:
        queue = ExecutorQueue(executor, "test", fail_on_first_exception=fail_on_first_exception)
        for idx in work:
            queue.submit_work(do_work, idx)

        try:
            queue.wait_for_submitted_work()
            return work_done, None
        except Exception as ex:
            return work_done, ex


def test_success_queue() -> None:
    work_done, ex = check_executor_queue(range(5), True)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is None
    work_done, ex = check_executor_queue(range(5), False)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is None


def test_fail_on_first() -> None:
    work_done, ex = check_executor_queue(range(10), True)
    assert work_done == [0, 1, 2, 3, 4]
    assert ex is not None
    assert ex.args[0] == "Abort 5"


def test_fail_late() -> None:
    work_done, ex = check_executor_queue(range(10), False)
    assert work_done == [0, 1, 2, 3, 4, 6, 7, 8, 9]
    assert ex is not None
    assert ex.args[0] == "Abort 5"



from typing import ClassVar, Dict, List, Optional, Type

from attrs import define, field

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resoto_plugin_aws.resource.kms import AwsKmsKey
from resoto_plugin_aws.resource.sns import AwsSnsTopic
from resotolib.baseresources import EdgeType, ModelReference
from resotolib.json_bender import S, Bend, Bender, ForallBend
from resotolib.types import Json


@define(eq=False, slots=False)
class AwsGlacierInventoryRetrievalParameters:
    kind: ClassVar[str] = "aws_glacier_job_inventory_retrieval_parameters"
    mapping: ClassVar[Dict[str, Bender]] = {
        "output_format": S("Format"),
        "start_date": S("StartDate"),
        "end_date": S("EndDate"),
        "limit": S("Limit"),
    }
    output_format: str = field(default=None)
    start_date: str = field(default=None)
    end_date: str = field(default=None)
    limit: str = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierSelectParameters:
    kind: ClassVar[str] = "aws_glacier_job_select_parameters"
    mapping: ClassVar[Dict[str, Bender]] = {
        "input_serialization": S("InputSerialization"),
        "expression_type": S("ExpressionType"),
        "expression": S("Expression"),
        "output_serialization": S("OutputSerialization"),
    }
    input_serialization: Dict[str, Dict[str, str]] = field(default=None)
    expression_type: str = field(default=None)
    expression: str = field(default=None)
    output_serialization: Dict[str, Dict[str, str]] = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierBucketEncryption:
    kind: ClassVar[str] = "aws_glacier_bucket_encryption"
    mapping: ClassVar[Dict[str, Bender]] = {
        "encryption_type": S("EncryptionType"),
        "kms_key_id": S("KMSKeyId"),
        "kms_context": S("KMSContext"),
    }
    encryption_type: str = field(default=None)
    kms_key_id: str = field(default=None)
    kms_context: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierAcl:
    kind: ClassVar[str] = "aws_glacier_acl"
    mapping: ClassVar[Dict[str, Bender]] = {
        "grantee": S("Grantee"),
        "permission": S("Permission"),
    }
    grantee: Dict[str, str] = field(default=None)
    permission: str = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierJobBucket:
    kind: ClassVar[str] = "aws_glacier_job_bucket"
    mapping: ClassVar[Dict[str, Bender]] = {
        "bucket_name": S("BucketName"),
        "prefix": S("Prefix"),
        "encryption": S("Encryption") >> Bend(AwsGlacierBucketEncryption.mapping),
        "canned_acl": S("CannedACL"),
        "access_control_list": S("AccessControlList") >> ForallBend(AwsGlacierAcl.mapping),
        "tagging": S("Tagging"),
        "user_metadata": S("UserMetadata"),
        "storage_class": S("StorageClass"),
    }
    bucket_name: str = field(default=None)
    prefix: str = field(default=None)
    encryption: AwsGlacierBucketEncryption = field(default=None)
    access_control_list: List[AwsGlacierAcl] = field(default=None)
    tagging: Optional[Dict[str, str]] = field(default=None)
    user_metadata: Dict[str, str] = field(default=None)
    storage_class: str = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierJobOutputLocation:
    kind: ClassVar[str] = "aws_glacier_job_output_location"
    mapping: ClassVar[Dict[str, Bender]] = {
        "s3": S("S3") >> Bend(AwsGlacierJobBucket.mapping),
    }
    s3: AwsGlacierJobBucket = field(default=None)


@define(eq=False, slots=False)
class AwsGlacierJob(AwsResource):
    kind: ClassVar[str] = "aws_glacier_job"
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {
            "delete": ["aws_kms_key"],
        },
        "successors": {"default": ["aws_kms_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("JobId"),
        "name": S("JobId"),
        "ctime": S("CreationDate"),
        "vault_arn": S("VaultARN"),
        "description": S("JobDescription"),
        "glacier_job_action": S("Action"),
        "glacier_job_archive_id": S("ArchiveId"),
        "glacier_job_vault_arn": S("VaultARN"),
        "glacier_job_completed": S("Completed"),
        "glacier_job_status_code": S("StatusCode"),
        "glacier_job_status_message": S("StatusMessage"),
        "glacier_job_archive_size_in_bytes": S("ArchiveSizeInBytes"),
        "glacier_job_inventory_size_in_bytes": S("InventorySizeInBytes"),
        "glacier_job_sns_topic": S("SNSTopic"),
        "glacier_job_completion_date": S("CompletionDate"),
        "glacier_job_sha256_tree_hash": S("SHA256TreeHash"),
        "glacier_job_archive_sha256_tree_hash": S("ArchiveSHA256TreeHash"),
        "glacier_job_retrieval_byte_range": S("RetrievalByteRange"),
        "glacier_job_tier": S("Tier"),
        "glacier_job_inventory_retrieval_parameters": S("InventoryRetrievalParameters")
        >> Bend(AwsGlacierInventoryRetrievalParameters.mapping),
        "glacier_job_output_path": S("JobOutputPath"),
        "glacier_job_select_parameters": S("SelectParameters") >> Bend(AwsGlacierSelectParameters.mapping),
        "glacier_job_output_location": S("OutputLocation") >> Bend(AwsGlacierJobOutputLocation.mapping),
    }
    description: Optional[str] = field(default=None)
    glacier_job_action: str = field(default=None)
    glacier_job_archive_id: Optional[str] = field(default=None)
    glacier_job_vault_arn: Optional[str] = field(default=None)
    glacier_job_completed: bool = field(default=None)
    glacier_job_status_code: str = field(default=None)
    glacier_job_status_message: str = field(default=None)
    glacier_job_archive_size_in_bytes: Optional[int] = field(default=None)
    glacier_job_inventory_size_in_bytes: Optional[int] = field(default=None)
    glacier_job_sns_topic: str = field(default=None)
    glacier_job_completion_date: Optional[str] = field(default=None)
    glacier_job_sha256_tree_hash: Optional[str] = field(default=None)
    glacier_job_archive_sha256_tree_hash: Optional[str] = field(default=None)
    glacier_job_retrieval_byte_range: Optional[str] = field(default=None)
    glacier_job_tier: str = field(default=None)
    glacier_job_inventory_retrieval_parameters: Optional[AwsGlacierInventoryRetrievalParameters] = field(default=None)
    glacier_job_output_path: str = field(default=None)
    glacier_job_select_parameters: Optional[AwsGlacierSelectParameters] = field(default=None)
    glacier_job_output_location: Optional[AwsGlacierJobOutputLocation] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.glacier_job_output_location:
            builder.dependant_node(
                self,
                clazz=AwsKmsKey,
                id=AwsKmsKey.normalise_id(self.glacier_job_output_location.s3.encryption.kms_key_id),
            )
        if self.glacier_job_sns_topic:
            builder.add_edge(self, clazz=AwsSnsTopic, arn=self.glacier_job_sns_topic)


@define(eq=False, slots=False)
class AwsGlacierVault(AwsResource):
    kind: ClassVar[str] = "aws_glacier_vault"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("glacier", "list-vaults", "VaultList")
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_glacier_job"],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("VaultName"),
        "name": S("VaultName"),
        "ctime": S("CreationDate"),
        "arn": S("VaultARN"),
        "glacier_last_inventory_date": S("LastInventoryDate"),
        "glacier_number_of_archives": S("NumberOfArchives"),
        "glacier_size_in_bytes": S("SizeInBytes"),
    }
    glacier_last_inventory_date: Optional[str] = field(default=None)
    glacier_number_of_archives: Optional[int] = field(default=None)
    glacier_size_in_bytes: Optional[int] = field(default=None)

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [
            cls.api_spec,
            AwsApiSpec(cls.api_spec.service, "list-tags-for-vault"),
            AwsApiSpec(cls.api_spec.service, "list-jobs"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(vault: AwsGlacierVault) -> None:
            tags = builder.client.get("glacier", "list-tags-for-vault", "Tags", vaultName=vault.name)
            if tags:
                vault.tags = tags

        for vault in json:
            vault_instance = cls.from_api(vault)
            builder.add_node(vault_instance, vault)
            builder.submit_work(add_tags, vault_instance)
            for job in builder.client.list("glacier", "list-jobs", "JobList", vaultName=vault_instance.name):
                job_instance = AwsGlacierJob.from_api(job)
                builder.add_node(job_instance, job)
                builder.add_edge(vault_instance, EdgeType.default, node=job_instance)

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            aws_service="glacier", action="add-tags-to-vault", result_name=None, vaultName=self.name, Tags={key: value}
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            aws_service="glacier", action="remove-tags-from-vault", result_name=None, vaultName=self.name, TagKeys=[key]
        )
        return True

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(aws_service="glacier", action="delete-vault", result_name=None, vaultName=self.name)
        return True

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("glacier", "add-tags-to-vault"),
            AwsApiSpec("glacier", "remove-tags-from-vault"),
            AwsApiSpec("glacier", "delete-vault"),
        ]


resources: List[Type[AwsResource]] = [AwsGlacierVault, AwsGlacierJob]



from types import SimpleNamespace
from typing import cast, Any

from attr import evolve

from resoto_plugin_aws.resource.base import AwsResource
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import GraphBuilder, AwsRegion
from resoto_plugin_aws.resource.ec2 import AwsEc2InstanceType, AwsEc2Vpc
from resoto_plugin_aws.resource.elbv2 import AwsAlb
from resoto_plugin_aws.resource.iam import AwsIamServerCertificate
from resoto_plugin_aws.resource.service_quotas import AwsServiceQuota, RegionalQuotas
from resotolib.baseresources import EdgeType
from test.resources import round_trip_for, build_graph


def test_service_quotas() -> None:
    first, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    assert len(builder.resources_of(AwsServiceQuota)) >= 13


def test_instance_type_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsEc2InstanceType.collect_resources(builder)
    for _, it in builder.global_instance_types.items():
        builder.add_node(it, {})
    expect_quotas(builder, 3)


def test_volume_type_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    for vt in ["gp2", "gp3", "standard", "io1", "io2", "sc1", "st1"]:
        builder.add_node(builder.volume_type(vt), {})  # type: ignore
    expect_quotas(builder, 7)


def test_vpc_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsEc2Vpc.collect_resources(builder)
    expect_quotas(builder, 3)


def test_alb_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsAlb.collect_resources(builder)
    expect_quotas(builder, 2)


def test_iam_server_certificate_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type", region_name="global")
    assert len(builder.resources_of(AwsServiceQuota)) == 1
    AwsIamServerCertificate.collect_resources(builder)
    expect_quotas(builder, 2)


def test_regional_matcher() -> None:
    builder = build_graph(AwsResource)
    assert len(builder.resources_of(AwsResource)) == 0  # empty graph

    # create 2 regions
    region1 = builder.add_node(AwsRegion(id="eu-central-1", tags={}, name="eu-central-1"))
    region2 = builder.add_node(AwsRegion(id="us-east-1", tags={}, name="us-east-1"))
    # add instance types
    t1 = builder.add_node(AwsEc2InstanceType(id="t1", name="t1"))
    t2 = builder.add_node(AwsEc2InstanceType(id="t2", name="t2"))
    t1._region = region1
    t2._region = region2
    # add service quotas
    sq1 = builder.add_node(AwsServiceQuota(quota=1, id="q1"))
    sq2 = builder.add_node(AwsServiceQuota(quota=2, id="q2"))
    sq1._region = region1
    sq2._region = region2

    # connect both quotas in graph
    sq1.connect_in_graph(builder, {"matcher": evolve(RegionalQuotas["ec2"][0], region=region1)})
    sq2.connect_in_graph(builder, {"matcher": evolve(RegionalQuotas["ec2"][0], region=region2)})

    # sq1 is only connected to t1
    assert list(builder.graph.successors(sq1, EdgeType.default)) == [t1]
    # sq2 is only connected to t2
    assert list(builder.graph.successors(sq2, EdgeType.default)) == [t2]


def expect_quotas(builder: GraphBuilder, quotas: int) -> None:
    for node, data in builder.graph.nodes(data=True):
        if isinstance(node, AwsServiceQuota):
            node.connect_in_graph(builder, data.get("source", {}))
    # make sure edges have been created
    assert builder.graph.number_of_edges() == quotas


def test_tagging() -> None:
    quota, _ = round_trip_for(AwsServiceQuota, "usage", "quota_type")

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag-resource"
        assert kwargs["ResourceARN"] == quota.arn
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag-resource"
        assert kwargs["ResourceARN"] == quota.arn
        assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    quota.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    quota.delete_resource_tag(client, "foo")



from typing import ClassVar, Dict, Optional, List
from typing import Type

from attrs import define

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.kms import AwsKmsKey
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import ModelReference
from resotolib.json_bender import Bender, S, Bend, bend
from resotolib.types import Json


@define(eq=False, slots=False)
class AwsAthenaEncryptionConfiguration:
    kind: ClassVar[str] = "aws_athena_encryption_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {"encryption_option": S("EncryptionOption"), "kms_key": S("KmsKey")}
    encryption_option: Optional[str] = None
    kms_key: Optional[str] = None


@define(eq=False, slots=False)
class AwsAthenaResultConfiguration:
    kind: ClassVar[str] = "aws_athena_result_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "output_location": S("OutputLocation"),
        "encryption_configuration": S("EncryptionConfiguration") >> Bend(AwsAthenaEncryptionConfiguration.mapping),
        "expected_bucket_owner": S("ExpectedBucketOwner"),
    }
    output_location: Optional[str] = None
    encryption_configuration: Optional[AwsAthenaEncryptionConfiguration] = None
    expected_bucket_owner: Optional[str] = None


@define(eq=False, slots=False)
class AwsAthenaEngineVersion:
    kind: ClassVar[str] = "aws_athena_engine_version"
    mapping: ClassVar[Dict[str, Bender]] = {
        "selected_engine_version": S("SelectedEngineVersion"),
        "effective_engine_version": S("EffectiveEngineVersion"),
    }
    selected_engine_version: Optional[str] = None
    effective_engine_version: Optional[str] = None


@define(eq=False, slots=False)
class AwsAthenaWorkGroupConfiguration:
    kind: ClassVar[str] = "aws_athena_work_group_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "result_configuration": S("ResultConfiguration") >> Bend(AwsAthenaResultConfiguration.mapping),
        "enforce_work_group_configuration": S("EnforceWorkGroupConfiguration"),
        "publish_cloud_watch_metrics_enabled": S("PublishCloudWatchMetricsEnabled"),
        "bytes_scanned_cutoff_per_query": S("BytesScannedCutoffPerQuery"),
        "requester_pays_enabled": S("RequesterPaysEnabled"),
        "engine_version": S("EngineVersion") >> Bend(AwsAthenaEngineVersion.mapping),
    }
    result_configuration: Optional[AwsAthenaResultConfiguration] = None
    enforce_work_group_configuration: Optional[bool] = None
    publish_cloud_watch_metrics_enabled: Optional[bool] = None
    bytes_scanned_cutoff_per_query: Optional[int] = None
    requester_pays_enabled: Optional[bool] = None
    engine_version: Optional[AwsAthenaEngineVersion] = None


@define(eq=False, slots=False)
class AwsAthenaWorkGroup(AwsResource):
    kind: ClassVar[str] = "aws_athena_work_group"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("athena", "list-work-groups", "WorkGroups")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Name"),
        "ctime": S("CreationTime"),
        "name": S("Name"),
        "workgroup_state": S("State"),
        "workgroup_configuration": S("Configuration") >> Bend(AwsAthenaWorkGroupConfiguration.mapping),
        "description": S("Description"),
    }
    workgroup_state: Optional[str] = None
    workgroup_configuration: Optional[AwsAthenaWorkGroupConfiguration] = None
    description: Optional[str] = None

    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"delete": ["aws_kms_key"]},
        "successors": {"default": ["aws_kms_key"]},
    }

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [cls.api_spec, AwsApiSpec("athena", "get-work-group"), AwsApiSpec("athena", "list-tags-for-resource")]

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("athena", "tag-resource"),
            AwsApiSpec("athena", "untag-resource"),
            AwsApiSpec("athena", "delete-work-group"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def fetch_workgroup(name: str) -> Optional[AwsAthenaWorkGroup]:
            result = builder.client.get(
                aws_service="athena", action="get-work-group", result_name="WorkGroup", WorkGroup=name
            )
            if result is None:
                return None

            workgroup = AwsAthenaWorkGroup.from_api(result)
            workgroup.set_arn(
                builder=builder,
                resource=f"workgroup/{workgroup.name}",
            )

            return workgroup

        def add_tags(data_catalog: AwsAthenaWorkGroup) -> None:
            tags = builder.client.list(
                "athena",
                "list-tags-for-resource",
                "Tags",
                ResourceARN=data_catalog.arn,
            )
            if tags:
                data_catalog.tags = bend(ToDict(), tags)

        for js in json:
            if (name := js.get("Name")) is not None and isinstance(name, str):
                wg = fetch_workgroup(name)
                if wg is not None:
                    builder.add_node(wg)
                    builder.submit_work(add_tags, wg)

    # noinspection PyUnboundLocalVariable
    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if (
            (wc := self.workgroup_configuration)
            and (rc := wc.result_configuration)
            and (ec := rc.encryption_configuration)
        ):
            if ec.kms_key:
                builder.dependant_node(from_node=self, clazz=AwsKmsKey, id=AwsKmsKey.normalise_id(ec.kms_key))

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="tag-resource",
            result_name=None,
            ResourceARN=self.arn,
            Tags=[{"Key": key, "Value": value}],
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="untag-resource",
            result_name=None,
            ResourceARN=self.arn,
            TagKeys=[key],
        )
        return True

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="delete-work-group",
            result_name=None,
            WorkGroup=self.name,
            RecursiveDeleteOption=True,
        )
        return True


@define(eq=False, slots=False)
class AwsAthenaDataCatalog(AwsResource):
    kind: ClassVar[str] = "aws_athena_data_catalog"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("athena", "list-data-catalogs", "DataCatalogsSummary")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Name"),
        "name": S("Name"),
        "description": S("Description"),
        "datacatalog_type": S("Type"),
        "datacatalog_parameters": S("Parameters"),
    }
    description: Optional[str] = None
    datacatalog_type: Optional[str] = None
    datacatalog_parameters: Optional[Dict[str, str]] = None

    @classmethod
    def called_collect_apis(cls) -> List[AwsApiSpec]:
        return [cls.api_spec, AwsApiSpec("athena", "get-data-catalog"), AwsApiSpec("athena", "list-tags-for-resource")]

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("athena", "tag-resource"),
            AwsApiSpec("athena", "untag-resource"),
            AwsApiSpec("athena", "delete-data-catalog"),
        ]

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def fetch_data_catalog(data_catalog_name: str) -> Optional[AwsAthenaDataCatalog]:
            result = builder.client.get(
                aws_service="athena",
                action="get-data-catalog",
                result_name="DataCatalog",
                Name=data_catalog_name,
            )
            if result is None:
                return None
            catalog = AwsAthenaDataCatalog.from_api(result)
            catalog.set_arn(builder=builder, resource=f"datacatalog/{catalog.name}")
            return catalog

        def add_tags(data_catalog: AwsAthenaDataCatalog) -> None:
            tags = builder.client.list(
                "athena",
                "list-tags-for-resource",
                None,
                ResourceARN=data_catalog.arn,
            )
            if tags:
                data_catalog.tags = bend(S("Tags", default=[]) >> ToDict(), tags[0])

        for js in json:
            # we filter out the default data catalog as it is not possible to do much with it
            if (name := js.get("CatalogName")) is not None and isinstance(name, str) and name != "AwsDataCatalog":
                catalog = fetch_data_catalog(name)
                if catalog is not None:
                    builder.add_node(catalog)
                    builder.submit_work(add_tags, catalog)

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="tag-resource",
            result_name=None,
            ResourceARN=self.arn,
            Tags=[{"Key": key, "Value": value}],
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            aws_service=self.api_spec.service,
            action="untag-resource",
            result_name=None,
            ResourceARN=self.arn,
            TagKeys=[key],
        )
        return True

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(aws_service="athena", action="delete-data-catalog", result_name=None, Name=self.name)
        return True


resources: List[Type[AwsResource]] = [
    AwsAthenaWorkGroup,
    AwsAthenaDataCatalog,
]



from collections import defaultdict
from typing import Dict, Any, cast
from types import SimpleNamespace

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.route53 import AwsRoute53Zone
from test.resources import round_trip_for


def test_hosted_zone() -> None:
    first, builder = round_trip_for(AwsRoute53Zone)
    type_count: Dict[str, int] = defaultdict(int)
    for node in builder.graph.nodes:
        type_count[node.kind] += 1
    assert type_count["aws_route53_zone"] == 3
    assert type_count["aws_route53_resource_record_set"] == 2
    assert type_count["aws_route53_resource_record"] == 5
    assert len(first.tags) == 1


def test_tagging_zones() -> None:
    zone, _ = round_trip_for(AwsRoute53Zone)

    def validate_update_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "change-tags-for-resource":
            assert kwargs["ResourceId"] == zone.id
            assert kwargs["AddTags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "change-tags-for-resource":
            assert kwargs["ResourceId"] == zone.id
            assert kwargs["RemoveTagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    zone.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    zone.delete_resource_tag(client, "foo")


def test_delete_zones() -> None:
    zone, _ = round_trip_for(AwsRoute53Zone)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-hosted-zone"
        assert kwargs["Id"] == zone.id

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    zone.delete_resource(client)



from collections import defaultdict
from types import SimpleNamespace
from typing import Any, Dict, cast
from resoto_plugin_aws.resource.ecs import AwsEcsCluster, AwsEcsTaskDefinition
from test.resources import round_trip_for
from resoto_plugin_aws.aws_client import AwsClient


def test_ecs_cluster() -> None:
    first, builder = round_trip_for(AwsEcsCluster)
    assert len(builder.resources_of(AwsEcsCluster)) == 1
    type_count: Dict[str, int] = defaultdict(int)
    for node in builder.graph.nodes:
        type_count[node.kind] += 1
    assert type_count["aws_ecs_container_instance"] == 2
    assert type_count["aws_ecs_service"] == 1
    assert type_count["aws_ecs_task"] == 1
    assert type_count["aws_ecs_capacity_provider"] == 1


def test_tagging() -> None:
    instance, _ = round_trip_for(AwsEcsCluster)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag-resource"
        assert kwargs["resourceArn"] == instance.arn
        assert kwargs["tags"] == [{"key": "foo", "value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag-resource"
        assert kwargs["resourceArn"] == instance.arn
        assert kwargs["tagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    instance.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    instance.delete_resource_tag(client, "foo")


def test_cluster_deletion() -> None:
    cluster, _ = round_trip_for(AwsEcsCluster)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-cluster"
        assert kwargs["cluster"] == cluster.arn

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    cluster.delete_resource(client)


def test_ecs_task_definition() -> None:
    first, builder = round_trip_for(AwsEcsTaskDefinition)
    assert len(builder.resources_of(AwsEcsTaskDefinition)) == 1
    assert first.family == "nginx-sample-stack"
    assert first.revision == 1
    assert first.id == "nginx-sample-stack:1"



import logging
from typing import ClassVar, Dict, List, Optional, Type

from attr import define, field

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resoto_plugin_aws.resource.iam import AwsIamServerCertificate
from resoto_plugin_aws.resource.lambda_ import AwsLambdaFunction
from resoto_plugin_aws.resource.s3 import AwsS3Bucket
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import ModelReference
from resotolib.graph import Graph
from resotolib.json_bender import S, Bend, Bender, ForallBend, bend
from resotolib.types import Json

log = logging.getLogger("resoto.plugins.aws")


class CloudFrontResource:
    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:  # type: ignore
        def add_tags(res: AwsResource) -> None:
            tags = builder.client.get("cloudfront", "list-tags-for-resource", "Tags", Resource=res.arn)
            if tags:
                res.tags = bend(ToDict(), tags["Items"])

        for js in json:
            instance = cls.from_api(js)
            builder.add_node(instance, js)
            if instance.arn:
                builder.submit_work(add_tags, instance)

    @staticmethod
    def delete_cloudfront_resource(client: AwsClient, resource: str, id: str) -> bool:
        description = client.get("cloudfront", f"get-{resource}", None, None, Id=id)
        if description:
            etag = description.get("ETag", None)
            if etag:
                client.call(
                    aws_service="cloudfront",
                    action=f"delete-{resource}",
                    result_name=None,
                    Id=id,
                    IfMatch=etag,
                )
                return True
        return False


class CloudFrontTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service="cloudfront",
                action="tag-resource",
                result_name=None,
                Resource=self.arn,
                Tags={"Items": [{"Key": key, "Value": value}]},
            )
            return True
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            client.call(
                aws_service="cloudfront",
                action="untag-resource",
                result_name=None,
                Resource=self.arn,
                TagKeys={"Items": [key]},
            )
            return True
        return False

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return [
            AwsApiSpec("cloudfront", "tag-resource"),
            AwsApiSpec("cloudfront", "untag-resource"),
        ]


@define(eq=False, slots=False)
class AwsCloudFrontOriginCustomHeader:
    kind: ClassVar[str] = "aws_cloudfront_origin_custom_header"
    mapping: ClassVar[Dict[str, Bender]] = {"header_name": S("HeaderName"), "header_value": S("HeaderValue")}
    header_name: Optional[str] = field(default=None)
    header_value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCustomOriginConfig:
    kind: ClassVar[str] = "aws_cloudfront_custom_origin_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "http_port": S("HTTPPort"),
        "https_port": S("HTTPSPort"),
        "origin_protocol_policy": S("OriginProtocolPolicy"),
        "origin_ssl_protocol": S("OriginSslProtocols", "Items", default=[]),
        "origin_read_timeout": S("OriginReadTimeout"),
        "origin_keepalive_timeout": S("OriginKeepaliveTimeout"),
    }
    http_port: Optional[int] = field(default=None)
    https_port: Optional[int] = field(default=None)
    origin_protocol_policy: Optional[str] = field(default=None)
    origin_ssl_protocol: List[str] = field(factory=list)
    origin_read_timeout: Optional[int] = field(default=None)
    origin_keepalive_timeout: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontOriginShield:
    kind: ClassVar[str] = "aws_cloudfront_origin_shield"
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("Enabled"), "origin_shield_region": S("OriginShieldRegion")}
    enabled: Optional[bool] = field(default=None)
    origin_shield_region: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontOrigin:
    kind: ClassVar[str] = "aws_cloudfront_origin"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "domain_name": S("DomainName"),
        "origin_path": S("OriginPath"),
        "custom_header": S("CustomHeaders", "Items", default=[]) >> ForallBend(AwsCloudFrontOriginCustomHeader.mapping),
        "s3_origin_config": S("S3OriginConfig", "OriginAccessIdentity"),
        "custom_origin_config": S("CustomOriginConfig") >> Bend(AwsCloudFrontCustomOriginConfig.mapping),
        "connection_attempts": S("ConnectionAttempts"),
        "connection_timeout": S("ConnectionTimeout"),
        "origin_shield": S("OriginShield") >> Bend(AwsCloudFrontOriginShield.mapping),
        "origin_access_control_id": S("OriginAccessControlId"),
    }
    id: Optional[str] = field(default=None)
    domain_name: Optional[str] = field(default=None)
    origin_path: Optional[str] = field(default=None)
    custom_header: List[AwsCloudFrontOriginCustomHeader] = field(factory=list)
    s3_origin_config: Optional[str] = field(default=None)
    custom_origin_config: Optional[AwsCloudFrontCustomOriginConfig] = field(default=None)
    connection_attempts: Optional[int] = field(default=None)
    connection_timeout: Optional[int] = field(default=None)
    origin_shield: Optional[AwsCloudFrontOriginShield] = field(default=None)
    origin_access_control_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontOriginGroupFailoverCriteria:
    kind: ClassVar[str] = "aws_cloudfront_origin_group_failover_criteria"
    mapping: ClassVar[Dict[str, Bender]] = {"status_codes": S("StatusCodes", "Items", default=[])}
    status_codes: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontOriginGroupMembers:
    kind: ClassVar[str] = "aws_cloudfront_origin_group_members"
    mapping: ClassVar[Dict[str, Bender]] = {
        "origin_id": S("OriginId"),
    }
    members: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontOriginGroup:
    kind: ClassVar[str] = "aws_cloudfront_origin_group"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "failover_criteria": S("FailoverCriteria") >> Bend(AwsCloudFrontOriginGroupFailoverCriteria.mapping),
        "members": S("Members", "Items", default=[]) >> ForallBend(AwsCloudFrontOriginGroupMembers.mapping),
    }
    id: Optional[str] = field(default=None)
    failover_criteria: Optional[AwsCloudFrontOriginGroupFailoverCriteria] = field(default=None)
    members: List[AwsCloudFrontOriginGroupMembers] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontLambdaFunctionAssociation:
    kind: ClassVar[str] = "aws_cloudfront_lambda_function_association"
    mapping: ClassVar[Dict[str, Bender]] = {
        "lambda_function_arn": S("LambdaFunctionARN"),
        "event_type": S("EventType"),
        "include_body": S("IncludeBody"),
    }
    lambda_function_arn: Optional[str] = field(default=None)
    event_type: Optional[str] = field(default=None)
    include_body: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontFunctionAssociation:
    kind: ClassVar[str] = "aws_cloudfront_function_association"
    mapping: ClassVar[Dict[str, Bender]] = {"function_arn": S("FunctionARN"), "event_type": S("EventType")}
    function_arn: Optional[str] = field(default=None)
    event_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCookiePreference:
    kind: ClassVar[str] = "aws_cloudfront_cookie_preference"
    mapping: ClassVar[Dict[str, Bender]] = {
        "forward": S("Forward"),
        "whitelisted_names": S("WhitelistedNames", "Items", default=[]),
    }
    forward: Optional[str] = field(default=None)
    whitelisted_names: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontForwardedValues:
    kind: ClassVar[str] = "aws_cloudfront_forwarded_values"
    mapping: ClassVar[Dict[str, Bender]] = {
        "query_string": S("QueryString"),
        "cookies": S("Cookies") >> Bend(AwsCloudFrontCookiePreference.mapping),
        "headers": S("Headers", "Items", default=[]),
        "query_string_cache_keys": S("QueryStringCacheKeys", "Items", default=[]),
    }
    query_string: Optional[bool] = field(default=None)
    cookies: Optional[AwsCloudFrontCookiePreference] = field(default=None)
    headers: List[str] = field(factory=list)
    query_string_cache_keys: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontDefaultCacheBehavior:
    kind: ClassVar[str] = "aws_cloudfront_default_cache_behavior"
    mapping: ClassVar[Dict[str, Bender]] = {
        "target_origin_id": S("TargetOriginId"),
        "trusted_signers": S("TrustedSigners", "Items", default=[]),
        "trusted_key_groups": S("TrustedKeyGroups", "Items", default=[]),
        "viewer_protocol_policy": S("ViewerProtocolPolicy"),
        "allowed_methods": S("AllowedMethods", "Items", default=[]),
        "smooth_streaming": S("SmoothStreaming"),
        "compress": S("Compress"),
        "lambda_function_associations": S("LambdaFunctionAssociations", "Items", default=[])
        >> ForallBend(AwsCloudFrontLambdaFunctionAssociation.mapping),
        "function_association": S("FunctionAssociations", "Items", default=[])
        >> ForallBend(AwsCloudFrontFunctionAssociation.mapping),
        "field_level_encryption_id": S("FieldLevelEncryptionId"),
        "realtime_log_config_arn": S("RealtimeLogConfigArn"),
        "cache_policy_id": S("CachePolicyId"),
        "origin_request_policy_id": S("OriginRequestPolicyId"),
        "response_headers_policy_id": S("ResponseHeadersPolicyId"),
        "forwarded_values": S("ForwardedValues") >> Bend(AwsCloudFrontForwardedValues.mapping),
        "min_ttl": S("MinTTL"),
        "default_ttl": S("DefaultTTL"),
        "max_ttl": S("MaxTTL"),
    }
    target_origin_id: Optional[str] = field(default=None)
    trusted_signers: List[str] = field(factory=list)
    trusted_key_groups: List[str] = field(factory=list)
    viewer_protocol_policy: Optional[str] = field(default=None)
    allowed_methods: List[str] = field(factory=list)
    smooth_streaming: Optional[bool] = field(default=None)
    compress: Optional[bool] = field(default=None)
    lambda_function_association: List[AwsCloudFrontLambdaFunctionAssociation] = field(factory=list)
    function_association: List[AwsCloudFrontFunctionAssociation] = field(factory=list)
    field_level_encryption_id: Optional[str] = field(default=None)
    realtime_log_config_arn: Optional[str] = field(default=None)
    cache_policy_id: Optional[str] = field(default=None)
    origin_request_policy_id: Optional[str] = field(default=None)
    response_headers_policy_id: Optional[str] = field(default=None)
    forwarded_values: Optional[AwsCloudFrontForwardedValues] = field(default=None)
    min_ttl: Optional[int] = field(default=None)
    default_ttl: Optional[int] = field(default=None)
    max_ttl: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCacheBehavior:
    kind: ClassVar[str] = "aws_cloudfront_cache_behavior"
    mapping: ClassVar[Dict[str, Bender]] = {
        "path_pattern": S("PathPattern"),
        "target_origin_id": S("TargetOriginId"),
        "trusted_signers": S("TrustedSigners", "Items", default=[]),
        "trusted_key_groups": S("TrustedKeyGroups", "Items", default=[]),
        "viewer_protocol_policy": S("ViewerProtocolPolicy"),
        "allowed_methods": S("AllowedMethods", "Items", default=[]),
        "smooth_streaming": S("SmoothStreaming"),
        "compress": S("Compress"),
        "lambda_function_association": S("LambdaFunctionAssociations", "Items", default=[])
        >> ForallBend(AwsCloudFrontLambdaFunctionAssociation.mapping),
        "function_association": S("FunctionAssociations", "Items", default=[])
        >> ForallBend(AwsCloudFrontFunctionAssociation.mapping),
        "field_level_encryption_id": S("FieldLevelEncryptionId"),
        "realtime_log_config_arn": S("RealtimeLogConfigArn"),
        "cache_policy_id": S("CachePolicyId"),
        "origin_request_policy_id": S("OriginRequestPolicyId"),
        "response_headers_policy_id": S("ResponseHeadersPolicyId"),
        "forwarded_values": S("ForwardedValues") >> Bend(AwsCloudFrontForwardedValues.mapping),
        "min_ttl": S("MinTTL"),
        "default_ttl": S("DefaultTTL"),
        "max_ttl": S("MaxTTL"),
    }
    path_pattern: Optional[str] = field(default=None)
    target_origin_id: Optional[str] = field(default=None)
    trusted_signers: List[str] = field(factory=list)
    trusted_key_groups: List[str] = field(factory=list)
    viewer_protocol_policy: Optional[str] = field(default=None)
    allowed_methods: List[str] = field(factory=list)
    smooth_streaming: Optional[bool] = field(default=None)
    compress: Optional[bool] = field(default=None)
    lambda_function_association: List[AwsCloudFrontLambdaFunctionAssociation] = field(factory=list)
    function_association: List[AwsCloudFrontFunctionAssociation] = field(factory=list)
    field_level_encryption_id: Optional[str] = field(default=None)
    realtime_log_config_arn: Optional[str] = field(default=None)
    cache_policy_id: Optional[str] = field(default=None)
    origin_request_policy_id: Optional[str] = field(default=None)
    response_headers_policy_id: Optional[str] = field(default=None)
    forwarded_values: Optional[AwsCloudFrontForwardedValues] = field(default=None)
    min_ttl: Optional[int] = field(default=None)
    default_ttl: Optional[int] = field(default=None)
    max_ttl: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCustomErrorResponse:
    kind: ClassVar[str] = "aws_cloudfront_custom_error_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "error_code": S("ErrorCode"),
        "response_page_path": S("ResponsePagePath"),
        "response_code": S("ResponseCode"),
        "error_caching_min_ttl": S("ErrorCachingMinTTL"),
    }
    error_code: Optional[int] = field(default=None)
    response_page_path: Optional[str] = field(default=None)
    response_code: Optional[str] = field(default=None)
    error_caching_min_ttl: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontViewerCertificate:
    kind: ClassVar[str] = "aws_cloudfront_viewer_certificate"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cloudfront_default_certificate": S("CloudFrontDefaultCertificate"),
        "iam_certificate_id": S("IAMCertificateId"),
        "acm_certificate_arn": S("ACMCertificateArn"),
        "ssl_support_method": S("SSLSupportMethod"),
        "minimum_protocol_version": S("MinimumProtocolVersion"),
        "certificate": S("Certificate"),
        "certificate_source": S("CertificateSource"),
    }
    cloudfront_default_certificate: Optional[bool] = field(default=None)
    iam_certificate_id: Optional[str] = field(default=None)
    acm_certificate_arn: Optional[str] = field(default=None)
    ssl_support_method: Optional[str] = field(default=None)
    minimum_protocol_version: Optional[str] = field(default=None)
    certificate: Optional[str] = field(default=None)
    certificate_source: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontRestrictions:
    kind: ClassVar[str] = "aws_cloudfront_restrictions"
    mapping: ClassVar[Dict[str, Bender]] = {"geo_restriction": S("GeoRestriction", "Items", default=[])}
    geo_restriction: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontAliasICPRecordal:
    kind: ClassVar[str] = "aws_cloudfront_alias_icp_recordal"
    mapping: ClassVar[Dict[str, Bender]] = {"cname": S("CNAME"), "icp_recordal_status": S("ICPRecordalStatus")}
    cname: Optional[str] = field(default=None)
    icp_recordal_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontDistribution(CloudFrontTaggable, CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_distribution"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudfront", "list-distributions", "DistributionList.Items")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"delete": ["aws_lambda_function"]},
        "successors": {
            "default": [
                "aws_lambda_function",
                "aws_iam_server_certificate",
                "aws_cloudfront_function",
                "aws_cloudfront_realtime_log_config",
                "aws_cloudfront_field_level_encryption_config",
                "aws_cloudfront_response_headers_policy",
                "aws_cloudfront_cache_policy",
                "aws_cloudfront_origin_access_control",
                "aws_s3_bucket",
            ]
        },
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "mtime": S("LastModifiedTime"),
        "arn": S("ARN"),
        "distribution_status": S("Status"),
        "distribution_domain_name": S("DomainName"),
        "distribution_aliases": S("Aliases", "Items", default=[]),
        "distribution_origin": S("Origins", "Items", default=[]) >> ForallBend(AwsCloudFrontOrigin.mapping),
        "distribution_origin_group": S("OriginGroups", "Items", default=[])
        >> ForallBend(AwsCloudFrontOriginGroup.mapping),
        "distribution_default_cache_behavior": S("DefaultCacheBehavior")
        >> Bend(AwsCloudFrontDefaultCacheBehavior.mapping),
        "distribution_cache_behavior": S("CacheBehaviors", "Items", default=[])
        >> ForallBend(AwsCloudFrontCacheBehavior.mapping),
        "distribution_custom_error_response": S("CustomErrorResponses", "Items", default=[])
        >> ForallBend(AwsCloudFrontCustomErrorResponse.mapping),
        "distribution_comment": S("Comment"),
        "distribution_price_class": S("PriceClass"),
        "distribution_enabled": S("Enabled"),
        "distribution_viewer_certificate": S("ViewerCertificate") >> Bend(AwsCloudFrontViewerCertificate.mapping),
        "distribution_restrictions": S("Restrictions") >> Bend(AwsCloudFrontRestrictions.mapping),
        "distribution_web_acl_id": S("WebACLId"),
        "distribution_http_version": S("HttpVersion"),
        "distribution_is_ipv6_enabled": S("IsIPV6Enabled"),
        "distribution_alias_icp_recordals": S("AliasICPRecordals", default=[])
        >> ForallBend(AwsCloudFrontAliasICPRecordal.mapping),
    }
    distribution_status: Optional[str] = field(default=None)
    distribution_domain_name: Optional[str] = field(default=None)
    distribution_aliases: List[str] = field(factory=list)
    distribution_origin: List[AwsCloudFrontOrigin] = field(factory=list)
    distribution_origin_group: List[AwsCloudFrontOriginGroup] = field(factory=list)
    distribution_default_cache_behavior: Optional[AwsCloudFrontDefaultCacheBehavior] = field(default=None)
    distribution_cache_behavior: List[AwsCloudFrontCacheBehavior] = field(factory=list)
    distribution_custom_error_response: List[AwsCloudFrontCustomErrorResponse] = field(factory=list)
    distribution_comment: Optional[str] = field(default=None)
    distribution_price_class: Optional[str] = field(default=None)
    distribution_enabled: Optional[bool] = field(default=None)
    distribution_viewer_certificate: Optional[AwsCloudFrontViewerCertificate] = field(default=None)
    distribution_restrictions: Optional[AwsCloudFrontRestrictions] = field(default=None)
    distribution_web_acl_id: Optional[str] = field(default=None)
    distribution_http_version: Optional[str] = field(default=None)
    distribution_is_ipv6_enabled: Optional[bool] = field(default=None)
    distribution_alias_icp_recordals: List[AwsCloudFrontAliasICPRecordal] = field(factory=list)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-distribution-config"),
            AwsApiSpec("cloudfront", "update-distribution"),
            AwsApiSpec("cloudfront", "get-distribution"),
            AwsApiSpec("cloudfront", "delete-distribution"),
        ]

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        # edges from default cache behaviour
        if dcb := self.distribution_default_cache_behavior:
            for a in dcb.lambda_function_association:
                builder.dependant_node(self, clazz=AwsLambdaFunction, arn=a.lambda_function_arn)
            for b in dcb.function_association:
                builder.add_edge(self, clazz=AwsCloudFrontFunction, arn=b.function_arn)
            if dcb.realtime_log_config_arn:
                builder.add_edge(self, clazz=AwsCloudFrontRealtimeLogConfig, arn=dcb.realtime_log_config_arn)
            if dcb.field_level_encryption_id:
                builder.add_edge(self, clazz=AwsCloudFrontFieldLevelEncryptionConfig, id=dcb.field_level_encryption_id)
            if dcb.response_headers_policy_id:
                builder.add_edge(self, clazz=AwsCloudFrontResponseHeadersPolicy, id=dcb.response_headers_policy_id)
            if dcb.cache_policy_id:
                builder.add_edge(self, clazz=AwsCloudFrontCachePolicy, id=dcb.cache_policy_id)

        # edges from other cache behaviours
        for cb_item in self.distribution_cache_behavior:
            for c in cb_item.lambda_function_association:
                builder.add_edge(self, clazz=AwsLambdaFunction, arn=c.lambda_function_arn)
            for d in cb_item.function_association:
                builder.add_edge(self, clazz=AwsCloudFrontFunction, arn=d.function_arn)
            if cb_item.field_level_encryption_id:
                builder.add_edge(
                    self, clazz=AwsCloudFrontFieldLevelEncryptionConfig, id=cb_item.field_level_encryption_id
                )
            if cb_item.realtime_log_config_arn:
                builder.add_edge(self, clazz=AwsCloudFrontRealtimeLogConfig, arn=cb_item.realtime_log_config_arn)
            if cb_item.cache_policy_id:
                builder.add_edge(self, clazz=AwsCloudFrontCachePolicy, id=cb_item.cache_policy_id)
            if cb_item.response_headers_policy_id:
                builder.add_edge(self, clazz=AwsCloudFrontResponseHeadersPolicy, id=cb_item.response_headers_policy_id)

        # other edges
        if self.distribution_origin:
            for entry in self.distribution_origin:
                builder.add_edge(self, clazz=AwsCloudFrontOriginAccessControl, id=entry.origin_access_control_id)
                builder.add_edge(self, clazz=AwsS3Bucket, name=entry.id)

        if self.distribution_viewer_certificate and self.distribution_viewer_certificate.iam_certificate_id:
            builder.add_edge(
                self, clazz=AwsIamServerCertificate, id=self.distribution_viewer_certificate.iam_certificate_id
            )

        # TODO edge to ACM certificate when applicable (via self.distribution_viewer_certificate.acm_certificate_arn)
        # TODO edge to Web Acl when applicable (via self.distribution_web_acl_id)

    def pre_delete_resource(self, client: AwsClient, graph: Graph) -> bool:
        dist_config = client.get("cloudfront", "get-distribution-config", None, None, Id=self.id)
        if dist_config:
            dist_config["DistributionConfig"]["Enabled"] = False
            dist_config["IfMatch"] = dist_config["ETag"]
            dist_config.pop("ETag")
            update = client.call(
                "cloudfront",
                "update-distribution",
                None,
                None,
                DistributionConfig=dist_config["DistributionConfig"],
                Id=self.id,
                IfMatch=dist_config["IfMatch"],
            )
            if update:
                return True
        return False

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "distribution", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontFunctionConfig:
    kind: ClassVar[str] = "aws_cloudfront_function_config"
    mapping: ClassVar[Dict[str, Bender]] = {"comment": S("Comment"), "runtime": S("Runtime")}
    comment: Optional[str] = field(default=None)
    runtime: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontFunction(CloudFrontTaggable, CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_function"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudfront", "list-functions", "FunctionList.Items")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Name"),
        "arn": S("FunctionMetadata", "FunctionARN", default=None),
        "name": S("Name"),
        "ctime": S("FunctionMetadata", "CreatedTime", default=None),
        "mtime": S("FunctionMetadata", "LastModifiedTime", default=None),
        "function_status": S("Status"),
        "function_stage": S("FunctionMetadata", "Stage", default=None),
        "function_config": S("FunctionConfig") >> Bend(AwsCloudFrontFunctionConfig.mapping),
    }
    function_status: Optional[str] = field(default=None)
    function_stage: Optional[str] = field(default=None)
    function_config: Optional[AwsCloudFrontFunctionConfig] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "describe-function"),
            AwsApiSpec("cloudfront", "delete-function"),
        ]

    def delete_resource(self, client: AwsClient) -> bool:
        description = client.get(
            self.api_spec.service, "describe-function", None, None, Name=self.name, Stage=self.function_stage
        )
        if description:
            etag = description.get("ETag", None)
            if etag:
                client.call(
                    aws_service=self.api_spec.service,
                    action="delete-function",
                    result_name=None,
                    Name=self.name,
                    IfMatch=etag,
                )
                return True
        return False


@define(eq=False, slots=False)
class AwsCloudFrontPublicKey(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_public_key"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudfront", "list-public-keys", "PublicKeyList.Items")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "name": S("Name"),
        "ctime": S("CreatedTime"),
        "public_key_encoded_key": S("EncodedKey"),
        "public_key_comment": S("Comment"),
    }
    public_key_encoded_key: Optional[str] = field(default=None)
    public_key_comment: Optional[str] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-public-key"),
            AwsApiSpec("cloudfront", "delete-public-key"),
        ]

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "public-key", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontKinesisStreamConfig:
    kind: ClassVar[str] = "aws_cloudfront_kinesis_stream_config"
    mapping: ClassVar[Dict[str, Bender]] = {"role_arn": S("RoleARN"), "stream_arn": S("StreamARN")}
    role_arn: Optional[str] = field(default=None)
    stream_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontEndPoint:
    kind: ClassVar[str] = "aws_cloudfront_end_point"
    mapping: ClassVar[Dict[str, Bender]] = {
        "stream_type": S("StreamType"),
        "kinesis_stream_config": S("KinesisStreamConfig") >> Bend(AwsCloudFrontKinesisStreamConfig.mapping),
    }
    stream_type: Optional[str] = field(default=None)
    kinesis_stream_config: Optional[AwsCloudFrontKinesisStreamConfig] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontRealtimeLogConfig(CloudFrontTaggable, CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_realtime_log_config"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudfront", "list-realtime-log-configs", "RealtimeLogConfigs.Items")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Name"),
        "name": S("Name"),
        "arn": S("ARN"),
        "realtime_log_sampling_rate": S("SamplingRate"),
        "realtime_log_end_points": S("EndPoints", default=[]) >> ForallBend(AwsCloudFrontEndPoint.mapping),
        "realtime_log_fields": S("Fields", default=[]),
    }
    realtime_log_sampling_rate: Optional[int] = field(default=None)
    realtime_log_end_points: List[AwsCloudFrontEndPoint] = field(factory=list)
    realtime_log_fields: List[str] = field(factory=list)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [AwsApiSpec("cloudfront", "delete-realtime-log-config")]

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            aws_service=self.api_spec.service, action="delete-realtime-log-config", result_name=None, Name=self.name
        )
        return True


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyCorsConfig:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_cors_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "access_control_allow_origins": S("AccessControlAllowOrigins", "Items", default=[]),
        "access_control_allow_headers": S("AccessControlAllowHeaders", "Items", default=[]),
        "access_control_allow_methods": S("AccessControlAllowMethods", "Items", default=[]),
        "access_control_allow_credentials": S("AccessControlAllowCredentials"),
        "access_control_expose_headers": S("AccessControlExposeHeaders", "Items", default=[]),
        "access_control_max_age_sec": S("AccessControlMaxAgeSec"),
        "origin_override": S("OriginOverride"),
    }
    access_control_allow_origins: List[str] = field(factory=list)
    access_control_allow_headers: List[str] = field(factory=list)
    access_control_allow_methods: List[str] = field(factory=list)
    access_control_allow_credentials: Optional[bool] = field(default=None)
    access_control_expose_headers: List[str] = field(factory=list)
    access_control_max_age_sec: Optional[int] = field(default=None)
    origin_override: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyXSSProtection:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_xss_protection"
    mapping: ClassVar[Dict[str, Bender]] = {
        "override": S("Override"),
        "protection": S("Protection"),
        "mode_block": S("ModeBlock"),
        "report_uri": S("ReportUri"),
    }
    override: Optional[bool] = field(default=None)
    protection: Optional[bool] = field(default=None)
    mode_block: Optional[bool] = field(default=None)
    report_uri: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyFrameOptions:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_frame_options"
    mapping: ClassVar[Dict[str, Bender]] = {"override": S("Override"), "frame_option": S("FrameOption")}
    override: Optional[bool] = field(default=None)
    frame_option: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyReferrerPolicy:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_referrer_policy"
    mapping: ClassVar[Dict[str, Bender]] = {"override": S("Override"), "referrer_policy": S("ReferrerPolicy")}
    override: Optional[bool] = field(default=None)
    referrer_policy: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyContentSecurityPolicy:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_content_security_policy"
    mapping: ClassVar[Dict[str, Bender]] = {
        "override": S("Override"),
        "content_security_policy": S("ContentSecurityPolicy"),
    }
    override: Optional[bool] = field(default=None)
    content_security_policy: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyStrictTransportSecurity:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_strict_transport_security"
    mapping: ClassVar[Dict[str, Bender]] = {
        "override": S("Override"),
        "include_subdomains": S("IncludeSubdomains"),
        "preload": S("Preload"),
        "access_control_max_age_sec": S("AccessControlMaxAgeSec"),
    }
    override: Optional[bool] = field(default=None)
    include_subdomains: Optional[bool] = field(default=None)
    preload: Optional[bool] = field(default=None)
    access_control_max_age_sec: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicySecurityHeadersConfig:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_security_headers_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "xss_protection": S("XSSProtection") >> Bend(AwsCloudFrontResponseHeadersPolicyXSSProtection.mapping),
        "frame_options": S("FrameOptions") >> Bend(AwsCloudFrontResponseHeadersPolicyFrameOptions.mapping),
        "referrer_policy": S("ReferrerPolicy") >> Bend(AwsCloudFrontResponseHeadersPolicyReferrerPolicy.mapping),
        "content_security_policy": S("ContentSecurityPolicy")
        >> Bend(AwsCloudFrontResponseHeadersPolicyContentSecurityPolicy.mapping),
        "content_type_options": S("ContentTypeOptions", "Override"),
        "strict_transport_security": S("StrictTransportSecurity")
        >> Bend(AwsCloudFrontResponseHeadersPolicyStrictTransportSecurity.mapping),
    }
    xss_protection: Optional[AwsCloudFrontResponseHeadersPolicyXSSProtection] = field(default=None)
    frame_options: Optional[AwsCloudFrontResponseHeadersPolicyFrameOptions] = field(default=None)
    referrer_policy: Optional[AwsCloudFrontResponseHeadersPolicyReferrerPolicy] = field(default=None)
    content_security_policy: Optional[AwsCloudFrontResponseHeadersPolicyContentSecurityPolicy] = field(default=None)
    content_type_options: Optional[bool] = field(default=None)
    strict_transport_security: Optional[AwsCloudFrontResponseHeadersPolicyStrictTransportSecurity] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyServerTimingHeadersConfig:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_server_timing_headers_config"
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("Enabled"), "sampling_rate": S("SamplingRate")}
    enabled: Optional[bool] = field(default=None)
    sampling_rate: Optional[float] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyCustomHeader:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_custom_header"
    mapping: ClassVar[Dict[str, Bender]] = {"header": S("Header"), "value": S("Value"), "override": S("Override")}
    header: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)
    override: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicyConfig:
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "comment": S("Comment"),
        "name": S("Name"),
        "cors_config": S("CorsConfig") >> Bend(AwsCloudFrontResponseHeadersPolicyCorsConfig.mapping),
        "security_headers_config": S("SecurityHeadersConfig")
        >> Bend(AwsCloudFrontResponseHeadersPolicySecurityHeadersConfig.mapping),
        "server_timing_headers_config": S("ServerTimingHeadersConfig")
        >> Bend(AwsCloudFrontResponseHeadersPolicyServerTimingHeadersConfig.mapping),
        "custom_headers_config": S("CustomHeadersConfig", "Items", default=[])
        >> ForallBend(AwsCloudFrontResponseHeadersPolicyCustomHeader.mapping),
    }
    comment: Optional[str] = field(default=None)
    name: str = field(default=None)
    cors_config: Optional[AwsCloudFrontResponseHeadersPolicyCorsConfig] = field(default=None)
    security_headers_config: Optional[AwsCloudFrontResponseHeadersPolicySecurityHeadersConfig] = field(default=None)
    server_timing_headers_config: Optional[AwsCloudFrontResponseHeadersPolicyServerTimingHeadersConfig] = field(
        default=None
    )
    custom_headers_config: List[AwsCloudFrontResponseHeadersPolicyCustomHeader] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontResponseHeadersPolicy(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_response_headers_policy"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "cloudfront", "list-response-headers-policies", "ResponseHeadersPolicyList.Items"
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("ResponseHeadersPolicy", "Id"),
        "mtime": S("ResponseHeadersPolicy", "LastModifiedTime"),
        "response_headers_policy_type": S("Type"),
        "response_headers_policy_config": S("ResponseHeadersPolicy", "ResponseHeadersPolicyConfig")
        >> Bend(AwsCloudFrontResponseHeadersPolicyConfig.mapping),
    }

    response_headers_policy_type: Optional[str] = field(default=None)
    response_headers_policy_config: Optional[AwsCloudFrontResponseHeadersPolicyConfig] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-response-headers-policy"),
            AwsApiSpec("cloudfront", "delete-response-headers-policy"),
        ]

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "response-headers-policy", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontS3Origin:
    kind: ClassVar[str] = "aws_cloudfront_s3_origin"
    mapping: ClassVar[Dict[str, Bender]] = {
        "domain_name": S("DomainName"),
        "origin_access_identity": S("OriginAccessIdentity"),
    }
    domain_name: Optional[str] = field(default=None)
    origin_access_identity: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontStreamingDistribution(CloudFrontTaggable, CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_streaming_distribution"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "cloudfront", "list-streaming-distributions", "StreamingDistributionList.Items"
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "mtime": S("LastModifiedTime"),
        "arn": S("ARN"),
        "streaming_distribution_status": S("Status"),
        "streaming_distribution_domain_name": S("DomainName"),
        "streaming_distribution_s3_origin": S("S3Origin") >> Bend(AwsCloudFrontS3Origin.mapping),
        "streaming_distribution_aliases": S("Aliases", "Items", default=[]),
        "streaming_distribution_trusted_signers": S("TrustedSigners", "Items", default=[]),
        "streaming_distribution_comment": S("Comment"),
        "streaming_distribution_price_class": S("PriceClass"),
        "streaming_distribution_enabled": S("Enabled"),
    }
    streaming_distribution_status: Optional[str] = field(default=None)
    streaming_distribution_domain_name: Optional[str] = field(default=None)
    streaming_distribution_trusted_signers: List[str] = field(factory=list)
    streaming_distribution_comment: Optional[str] = field(default=None)
    streaming_distribution_price_class: Optional[str] = field(default=None)
    streaming_distribution_enabled: Optional[bool] = field(default=None)

    # deleting streaming distributions is a multistep process:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudfront.html#CloudFront.Client.delete_streaming_distribution


@define(eq=False, slots=False)
class AwsCloudFrontOriginAccessControl(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_origin_access_control"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "cloudfront", "list-origin-access-controls", "OriginAccessControlList.Items"
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "name": S("Name"),
        "origin_access_control_description": S("Description"),
        "origin_access_control_signing_protocol": S("SigningProtocol"),
        "origin_access_control_signing_behavior": S("SigningBehavior"),
        "origin_access_control_origin_access_control_origin_type": S("OriginAccessControlOriginType"),
    }
    origin_access_control_description: Optional[str] = field(default=None)
    origin_access_control_signing_protocol: Optional[str] = field(default=None)
    origin_access_control_signing_behavior: Optional[str] = field(default=None)
    origin_access_control_origin_access_control_origin_type: Optional[str] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-origin-access-control"),
            AwsApiSpec("cloudfront", "delete-origin-access-control"),
        ]

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "origin-access-control", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontCachePolicyHeadersConfig:
    kind: ClassVar[str] = "aws_cloudfront_cache_policy_headers_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "header_behavior": S("HeaderBehavior"),
        "headers": S("Headers", "Items", default=[]),
    }
    header_behavior: Optional[str] = field(default=None)
    headers: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontCachePolicyCookiesConfig:
    kind: ClassVar[str] = "aws_cloudfront_cache_policy_cookies_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cookie_behavior": S("CookieBehavior"),
        "cookies": S("Cookies", default=[]),
    }
    cookie_behavior: Optional[str] = field(default=None)
    cookies: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontCachePolicyQueryStringsConfig:
    kind: ClassVar[str] = "aws_cloudfront_cache_policy_query_strings_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "query_string_behavior": S("QueryStringBehavior"),
        "query_strings": S("QueryStrings", "Items", default=[]),
    }
    query_string_behavior: Optional[str] = field(default=None)
    query_strings: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontParametersInCacheKeyAndForwardedToOrigin:
    kind: ClassVar[str] = "aws_cloudfront_parameters_in_cache_key_and_forwarded_to_origin"
    mapping: ClassVar[Dict[str, Bender]] = {
        "enable_accept_encoding_gzip": S("EnableAcceptEncodingGzip"),
        "enable_accept_encoding_brotli": S("EnableAcceptEncodingBrotli"),
        "headers_config": S("HeadersConfig") >> Bend(AwsCloudFrontCachePolicyHeadersConfig.mapping),
        "cookies_config": S("CookiesConfig") >> Bend(AwsCloudFrontCachePolicyCookiesConfig.mapping),
        "query_strings_config": S("QueryStringsConfig") >> Bend(AwsCloudFrontCachePolicyQueryStringsConfig.mapping),
    }
    enable_accept_encoding_gzip: Optional[bool] = field(default=None)
    enable_accept_encoding_brotli: Optional[bool] = field(default=None)
    headers_config: Optional[AwsCloudFrontCachePolicyHeadersConfig] = field(default=None)
    cookies_config: Optional[AwsCloudFrontCachePolicyCookiesConfig] = field(default=None)
    query_strings_config: Optional[AwsCloudFrontCachePolicyQueryStringsConfig] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCachePolicyConfig:
    kind: ClassVar[str] = "aws_cloudfront_cache_policy_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "comment": S("Comment"),
        "name": S("Name"),
        "default_ttl": S("DefaultTTL"),
        "max_ttl": S("MaxTTL"),
        "min_ttl": S("MinTTL"),
        "parameters_in_cache_key_and_forwarded_to_origin": S("ParametersInCacheKeyAndForwardedToOrigin")
        >> Bend(AwsCloudFrontParametersInCacheKeyAndForwardedToOrigin.mapping),
    }
    comment: Optional[str] = field(default=None)
    name: str = field(default=None)
    default_ttl: Optional[int] = field(default=None)
    max_ttl: Optional[int] = field(default=None)
    min_ttl: Optional[int] = field(default=None)
    parameters_in_cache_key_and_forwarded_to_origin: Optional[
        AwsCloudFrontParametersInCacheKeyAndForwardedToOrigin
    ] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontCachePolicy(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_cache_policy"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudfront", "list-cache-policies", "CachePolicyList.Items")
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("CachePolicy", "Id"),
        "name": S("CachePolicy", "CachePolicyConfig", "Name"),
        "mtime": S("CachePolicy", "LastModifiedTime"),
        "cache_policy_config": S("CachePolicy", "CachePolicyConfig") >> Bend(AwsCloudFrontCachePolicyConfig.mapping),
    }
    cache_policy_config: Optional[AwsCloudFrontCachePolicyConfig] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-cache-policy"),
            AwsApiSpec("cloudfront", "delete-cache-policy"),
        ]

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "cache-policy", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontQueryArgProfile:
    kind: ClassVar[str] = "aws_cloudfront_query_arg_profile"
    mapping: ClassVar[Dict[str, Bender]] = {"query_arg": S("QueryArg"), "profile_id": S("ProfileId")}
    query_arg: Optional[str] = field(default=None)
    profile_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontQueryArgProfileConfig:
    kind: ClassVar[str] = "aws_cloudfront_query_arg_profile_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "forward_when_query_arg_profile_is_unknown": S("ForwardWhenQueryArgProfileIsUnknown"),
        "query_arg_profiles": S("QueryArgProfiles", "Items", default=[])
        >> ForallBend(AwsCloudFrontQueryArgProfile.mapping),
    }
    forward_when_query_arg_profile_is_unknown: Optional[bool] = field(default=None)
    query_arg_profiles: List[AwsCloudFrontQueryArgProfile] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontContentTypeProfile:
    kind: ClassVar[str] = "aws_cloudfront_content_type_profile"
    mapping: ClassVar[Dict[str, Bender]] = {
        "format": S("Format"),
        "profile_id": S("ProfileId"),
        "content_type": S("ContentType"),
    }
    format: Optional[str] = field(default=None)
    profile_id: Optional[str] = field(default=None)
    content_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudFrontContentTypeProfileConfig:
    kind: ClassVar[str] = "aws_cloudfront_content_type_profile_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "forward_when_content_type_is_unknown": S("ForwardWhenContentTypeIsUnknown"),
        "content_type_profiles": S("ContentTypeProfiles", "Items", default=[])
        >> ForallBend(AwsCloudFrontContentTypeProfile.mapping),
    }
    forward_when_content_type_is_unknown: Optional[bool] = field(default=None)
    content_type_profiles: List[AwsCloudFrontContentTypeProfile] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontFieldLevelEncryptionConfig(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_field_level_encryption_config"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "cloudfront", "list-field-level-encryption-configs", "FieldLevelEncryptionList.Items"
    )
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_cloudfront_field_level_encryption_profile"]}
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "mtime": S("LastModifiedTime"),
        "field_level_encryption_config_comment": S("Comment"),
        "field_level_encryption_config_query_arg_profile_config": S("QueryArgProfileConfig")
        >> Bend(AwsCloudFrontQueryArgProfileConfig.mapping),
        "field_level_encryption_config_content_type_profile_config": S("ContentTypeProfileConfig")
        >> Bend(AwsCloudFrontContentTypeProfileConfig.mapping),
    }
    field_level_encryption_config_comment: Optional[str] = field(default=None)
    field_level_encryption_config_query_arg_profile_config: Optional[AwsCloudFrontQueryArgProfileConfig] = field(
        default=None
    )
    field_level_encryption_config_content_type_profile_config: Optional[AwsCloudFrontContentTypeProfileConfig] = field(
        default=None
    )

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-field-level-encryption-config"),
            AwsApiSpec("cloudfront", "delete-field-level-encryption-config"),
        ]

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if (
            self.field_level_encryption_config_content_type_profile_config
            and self.field_level_encryption_config_content_type_profile_config.content_type_profiles
        ):
            for entry in self.field_level_encryption_config_content_type_profile_config.content_type_profiles:
                builder.add_edge(
                    self,
                    clazz=AwsCloudFrontFieldLevelEncryptionProfile,
                    id=entry.profile_id,
                )

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "field-level-encryption-config", self.id)


@define(eq=False, slots=False)
class AwsCloudFrontEncryptionEntity:
    kind: ClassVar[str] = "aws_cloudfront_encryption_entity"
    mapping: ClassVar[Dict[str, Bender]] = {
        "public_key_id": S("PublicKeyId"),
        "provider_id": S("ProviderId"),
        "field_patterns": S("FieldPatterns", "Items", default=[]),
    }
    public_key_id: Optional[str] = field(default=None)
    provider_id: Optional[str] = field(default=None)
    field_patterns: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudFrontFieldLevelEncryptionProfile(CloudFrontResource, AwsResource):
    kind: ClassVar[str] = "aws_cloudfront_field_level_encryption_profile"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec(
        "cloudfront", "list-field-level-encryption-profiles", "FieldLevelEncryptionProfileList.Items"
    )
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["aws_cloudfront_public_key"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "name": S("Name"),
        "mtime": S("LastModifiedTime"),
        "field_level_encryption_profile_encryption_entities": S("EncryptionEntities", "Items", default=[])
        >> ForallBend(AwsCloudFrontEncryptionEntity.mapping),
        "field_level_encryption_profile_comment": S("Comment"),
    }
    field_level_encryption_profile_encryption_entities: List[AwsCloudFrontEncryptionEntity] = field(factory=list)
    field_level_encryption_profile_comment: Optional[str] = field(default=None)

    @classmethod
    def called_mutator_apis(cls) -> List[AwsApiSpec]:
        return super().called_mutator_apis() + [
            AwsApiSpec("cloudfront", "get-field-level-encryption-profile"),
            AwsApiSpec("cloudfront", "delete-field-level-encryption-profile"),
        ]

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.field_level_encryption_profile_encryption_entities:
            for entry in self.field_level_encryption_profile_encryption_entities:
                builder.add_edge(
                    self,
                    clazz=AwsCloudFrontPublicKey,
                    id=entry.public_key_id,
                )

    def delete_resource(self, client: AwsClient) -> bool:
        return self.delete_cloudfront_resource(client, "field-level-encryption-profile", self.id)


resources: List[Type[AwsResource]] = [
    AwsCloudFrontDistribution,
    AwsCloudFrontFunction,
    AwsCloudFrontPublicKey,
    AwsCloudFrontRealtimeLogConfig,
    AwsCloudFrontResponseHeadersPolicy,
    AwsCloudFrontStreamingDistribution,
    AwsCloudFrontOriginAccessControl,
    AwsCloudFrontCachePolicy,
    AwsCloudFrontFieldLevelEncryptionConfig,
    AwsCloudFrontFieldLevelEncryptionProfile,
]



import logging
import multiprocessing
from concurrent import futures
from typing import List, Optional, Union, Any, Dict

import botocore.exceptions
from botocore.model import OperationModel, Shape, StringShape, ListShape, StructureShape
from jsons import pascalcase
from prometheus_client import Summary, Counter

import resotolib.logger
import resotolib.proc
from resoto_plugin_aws.aws_client import AwsClient
from resotolib.args import ArgumentParser
from resotolib.args import Namespace
from resotolib.baseplugin import BaseCollectorPlugin
from resotolib.baseresources import (
    BaseResource,
    metrics_resource_cleanup_exceptions,
    BaseAccount,
    BaseRegion,
    metrics_resource_pre_cleanup_exceptions,
)
from resotolib.baseresources import Cloud
from resotolib.config import Config, RunningConfig, current_config
from resotolib.core.actions import CoreFeedback
from resotolib.core.custom_command import execute_command_on_resource
from resotolib.graph import Graph
from resotolib.logger import log, setup_logger
from resotolib.types import JsonElement
from resotolib.utils import log_runtime, NoExitArgumentParser
from .collector import AwsAccountCollector
from .configuration import AwsConfig
from .resource.base import AwsAccount, AwsResource, get_client
from .utils import aws_session

logging.getLogger("boto").setLevel(logging.CRITICAL)

metrics_collect = Summary("resoto_plugin_aws_collect_seconds", "Time it took the collect() method")
metrics_unhandled_account_exceptions = Counter(
    "resoto_plugin_aws_unhandled_account_exceptions_total",
    "Unhandled AWS Plugin Account Exceptions",
    ["account"],
)


class AWSCollectorPlugin(BaseCollectorPlugin):
    cloud = "aws"

    def __init__(self) -> None:
        super().__init__()
        self.__regions: List[str] = []
        self.core_feedback: Optional[CoreFeedback] = None

    @staticmethod
    def add_config(cfg: Config) -> None:
        cfg.add_config(AwsConfig)

    @metrics_collect.time()  # type: ignore
    def collect(self) -> None:
        log.debug("plugin: AWS collecting resources")
        assert self.core_feedback, "core_feedback is not set"
        cloud = Cloud(id=self.cloud, name="AWS")

        accounts = get_accounts(self.core_feedback.with_context(cloud.id))
        if len(accounts) == 0:
            log.error("No accounts found")
            return
        for account in accounts:
            add_str = ""
            if account.role:
                add_str += f" role {account.role}"
            if account.profile:
                add_str += f" profile {account.profile}"
            log.debug(f"Found {account.rtdname}{add_str}")

        max_workers = len(accounts) if len(accounts) < Config.aws.account_pool_size else Config.aws.account_pool_size
        pool_args = {"max_workers": max_workers}
        if Config.aws.fork_process:
            pool_args["mp_context"] = multiprocessing.get_context("spawn")
            pool_args["initializer"] = resotolib.proc.initializer
            pool_executor = futures.ProcessPoolExecutor
        else:
            pool_executor = futures.ThreadPoolExecutor  # type: ignore

        with pool_executor(**pool_args) as executor:
            wait_for = [
                executor.submit(
                    collect_account,
                    account,
                    self.regions(profile=account.profile),
                    ArgumentParser.args,
                    Config.running_config,
                    self.core_feedback.with_context(cloud.id, account.dname),
                    cloud,
                )
                for account in accounts
            ]
            for future in futures.as_completed(wait_for):
                account_graph = future.result()
                if not isinstance(account_graph, Graph):
                    log.debug(f"Skipping account graph of invalid type {type(account_graph)}")
                    continue
                self.graph.merge(account_graph)

    def regions(self, profile: Optional[str] = None) -> List[str]:
        if len(self.__regions) == 0:
            if not Config.aws.region or (isinstance(Config.aws.region, list) and len(Config.aws.region) == 0):
                log.debug("AWS region not specified, assuming all regions")
                self.__regions = all_regions(profile=profile)
            else:
                self.__regions = list(Config.aws.region)
        return self.__regions

    @execute_command_on_resource(
        name="aws",
        info="Execute aws commands on AWS resources",
        args_description={
            "--account": "[Optional] The AWS account identifier.",
            "--role": "[Optional] The AWS role.",
            "--profile": "[Optional] The AWS profile to use.",
            "--region": "[Optional] The AWS region.",
            "service": "Defines the AWS service, like ec2, s3, iam, etc.",
            "operation": "Defines the operation to execute.",
            "operation_args": "Defines the arguments for the operation. The parameters depend on the operation.",
        },
        description="Execute an operation on an AWS resource.\n"
        "For a list of services with respective operations see "
        "https://awscli.amazonaws.com/v2/documentation/api/latest/reference/index.html#available-services\n\n"
        "By default the operation runs with the same credentials as the collect process.\n"
        "You can override the credentials by providing the account, role, profile and region arguments.\n\n"
        "There are two modes of operation:\n"
        "1. Use a search and then pipe the result of the search into the `aws` command. "
        "Every resource matched by the search will invoke this command. "
        "You can use templating parameter to define the exact invocation arguments.\n"
        "2. Call the `aws` command directly without passing any resource to interact "
        "with AWS using the credentials defined via configuration.\n\n"
        "## Examples\n\n"
        "```shell\n"
        "# Search for all ec2 volumes and then call describe-volumes on each volume\n"
        "# See AWS CLI for available services and commands.\n"
        "# Please note the {id} parameter. The aws command is invoked for any volume and replaces the id parameter"
        " with the volume id.\n"
        "> search is(aws_ec2_volume) | aws ec2 describe-volume-attribute --volume-id {id} --attribute autoEnableIO\n"
        "AutoEnableIO:\n"
        "Value: false\n"
        "VolumeId: vol-009b0a28d2754927e\n\n"
        "# Get the current caller identity\n"
        "> aws sts get-caller-identity\n"
        "UserId: AIDA42373XXXXXXXXXXXX\n"
        "Account: '882374444444'\n"
        "Arn: arn:aws:iam::882374444444:user/matthias\n"
        "```\n\n",
        allowed_on_kind="aws_resource",
    )
    def call_aws_function(
        self, config: Config, resource: Optional[BaseResource], args: List[str]
    ) -> Union[JsonElement, BaseResource]:
        parser = NoExitArgumentParser()
        parser.add_argument("--account")
        parser.add_argument("--role")
        parser.add_argument("--profile")
        parser.add_argument("--region")
        parser.add_argument("service")
        parser.add_argument("operation")
        p, remaining = parser.parse_known_args(args)
        cfg = config.aws

        def adjust_shape(o: str, shape: Optional[Shape]) -> Any:
            if shape is None or isinstance(shape, StringShape):
                return o
            elif isinstance(shape, ListShape):
                return o.split(",")
            else:
                # map and structure types are currently not supported
                raise ValueError(f"Cannot convert {o} to {shape}")

        def coerce_args(fn_args: List[str], om: OperationModel) -> Dict[str, Any]:
            members: Dict[str, Shape] = om.input_shape.members if isinstance(om.input_shape, StructureShape) else {}
            param_name: Optional[str] = None
            param_shape: Optional[Shape] = None
            arg_val: Dict[str, Any] = {}
            for arg in fn_args:
                if arg.startswith("--"):
                    name = pascalcase(arg.removeprefix("--"))
                    param_name = name
                    param_shape = members.get(name)
                    bool_value = True
                    if param_shape is None and arg.startswith("--no-"):
                        name = name[2:]
                        param_name = name
                        param_shape = members.get(name)
                        bool_value = False
                    if param_shape is None:
                        raise ValueError(f"AWS: Unknown parameter {arg}")
                    if param_shape.name == "Boolean" or param_shape.type_name == "Boolean":
                        arg_val[name] = bool_value
                        param_shape = None
                        param_name = None
                elif param_name is not None:
                    arg_val[param_name] = adjust_shape(arg, param_shape)
                    param_name = None
                    param_shape = None
                else:
                    raise ValueError(f"AWS: Unexpected argument {arg}")
            return arg_val

        def create_client() -> AwsClient:
            role = p.role or cfg.role
            region = p.region or (cfg.region[0] if cfg.region else None)
            profile = p.profile or (cfg.profiles[0] if cfg.profiles else None)
            # possibly expensive call: account id is looked up if not provided
            account = p.account or (cfg.account[0] if cfg.account else current_account_id(profile))
            return AwsClient(cfg, account, role=role, profile=profile, region=region)

        client = get_client(current_config(), resource) if resource else create_client()

        # try to get the output shape of the operation
        service_model = client.service_model(p.service)
        op: OperationModel = service_model.operation_model(pascalcase(p.operation))
        output_shape = op.output_shape.type_name
        func_args = coerce_args(remaining, op)

        # call single and treat result as list
        response = client.call_single(p.service, p.operation, None, **func_args)
        if response is None:
            result = []
        elif isinstance(response, list):
            result = response
        else:
            result = [response]

        # Remove the "ResponseMetadata" from the result
        for elem in result:
            if isinstance(elem, dict):
                elem.pop("ResponseMetadata", None)
        return (result[0] if len(result) == 1 else None) if output_shape == "structure" else result

    @staticmethod
    def update_tag(config: Config, resource: BaseResource, key: str, value: str) -> bool:
        """Update the tag of a resource"""
        if isinstance(resource, AwsResource):
            client = get_client(config, resource)
            return resource.update_resource_tag(client, key, value)

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def delete_tag(config: Config, resource: BaseResource, key: str) -> bool:
        """Delete the tag of a resource"""
        if isinstance(resource, AwsResource):
            client = get_client(config, resource)
            return resource.delete_resource_tag(client, key)

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def pre_cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        if isinstance(resource, AwsResource):
            client = get_client(config, resource)

            if not hasattr(resource, "pre_delete_resource"):
                return True

            if graph is None:
                graph = resource._graph

            if resource.protected:
                log.error(f"Resource {resource.rtdname} is protected - refusing modification")
                resource.log("Modification was requested even though resource is protected" " - refusing")
                return False

            if resource.phantom:
                log.error(f"Can't cleanup phantom resource {resource.rtdname}")
                return False

            if resource.cleaned:
                log.debug(f"Resource {resource.rtdname} has already been cleaned up")
                return True

            account = resource.account(graph)
            region = resource.region(graph)
            if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
                log.error(f"Could not determine account or region for pre cleanup of {resource.rtdname}")
                return False

            log_suffix = f" in account {account.dname} region {region.name}"
            resource.log("Trying to run pre clean up")
            log.debug(f"Trying to run pre clean up {resource.rtdname}{log_suffix}")
            try:
                if not getattr(resource, "pre_delete_resource")(client, graph):
                    resource.log("Failed to run pre clean up")
                    log.error(f"Failed to run pre clean up {resource.rtdname}{log_suffix}")
                    return False
                resource.log("Successfully ran pre clean up")
                log.info(f"Successfully ran pre clean up {resource.rtdname}{log_suffix}")
            except Exception as e:
                resource.log("An error occurred during pre clean up", exception=e)
                log.exception(f"An error occurred during pre clean up {resource.rtdname}{log_suffix}")
                cloud = resource.cloud(graph)
                metrics_resource_pre_cleanup_exceptions.labels(
                    cloud=cloud.name,
                    account=account.dname,
                    region=region.name,
                    kind=resource.kind,
                ).inc()
                raise
            return True

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        if isinstance(resource, AwsResource):

            client = get_client(config, resource)

            if resource.phantom:
                raise RuntimeError(f"Can't cleanup phantom resource {resource.rtdname}")

            if resource.cleaned:
                log.debug(f"Resource {resource.rtdname} has already been cleaned up")
                return True

            if resource.protected:
                log.error(f"Resource {resource.rtdname} is protected - refusing modification")
                resource.log(("Modification was requested even though resource is protected" " - refusing"))
                return False

            resource._changes.add("cleaned")
            if graph is None:
                graph = resource._graph

            account = resource.account(graph)
            region = resource.region(graph)
            if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
                raise RuntimeError(f"Could not determine account or region for cleanup of {resource.rtdname}")

            log_suffix = f" in account {account.dname} region {region.name}"
            resource.log("Trying to clean up")
            log.debug(f"Trying to clean up {resource.rtdname}{log_suffix}")
            try:
                if deleted := resource.delete_resource(client):
                    resource._cleaned = True
                    resource.log("Successfully cleaned up")
                    log.info(f"Successfully cleaned up {resource.rtdname}{log_suffix}")
            except Exception as e:
                resource.log("An error occurred during clean up", exception=e)
                log.exception(f"An error occurred during clean up {resource.rtdname}{log_suffix}")
                cloud = resource.cloud(graph)
                metrics_resource_cleanup_exceptions.labels(
                    cloud=cloud.name,
                    account=account.dname,
                    region=region.name,
                    kind=resource.kind,
                ).inc()
                return False
            if not deleted:
                raise RuntimeError(f"Failed to clean up {resource.rtdname}{log_suffix}")
            return True

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")


def authenticated(account: AwsAccount, core_feedback: CoreFeedback) -> bool:
    try:
        log.debug(f"AWS testing credentials for {account.rtdname}")
        session = aws_session(account.id, account.role, account.profile)
        _ = session.client("sts").get_caller_identity().get("Account")
    except botocore.exceptions.NoCredentialsError:
        core_feedback.error(f"No AWS credentials found for {account.rtdname}", log)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AuthFailure":
            core_feedback.error(f"Unable to validate the provided access credentials for {account.rtdname}", log)
        elif e.response["Error"]["Code"] == "InvalidClientTokenId":
            core_feedback.error(f"Unable to validate the provided security token for {account.rtdname}", log)
        elif e.response["Error"]["Code"] == "ExpiredToken":
            core_feedback.error(f"Security token included in the request is expired for {account.rtdname}", log)
        elif e.response["Error"]["Code"] == "AccessDenied":
            core_feedback.error(f"Access Denied to {account.rtdname}: {e}", log)
        else:
            raise
        return False
    return True


def current_account_id(profile: Optional[str] = None) -> str:
    session = aws_session(profile=profile)
    return session.client("sts").get_caller_identity().get("Account")  # type: ignore


def set_account_names(accounts: List[AwsAccount]) -> None:
    def set_account_name(account: AwsAccount) -> None:
        try:
            account_aliases = (
                aws_session(account.id, account.role, account.profile)
                .client("iam")
                .list_account_aliases()
                .get("AccountAliases", [])
            )
            if len(account_aliases) > 0:
                account.name = account_aliases[0]
        except Exception:
            pass

    if len(accounts) == 0:
        return

    max_workers = len(accounts) if len(accounts) < Config.aws.account_pool_size else Config.aws.account_pool_size
    with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        executor.map(set_account_name, accounts)


def get_accounts(core_feedback: CoreFeedback) -> List[AwsAccount]:
    accounts = []
    profiles = [None]

    if Config.aws.assume_current and not Config.aws.do_not_scrape_current:
        msg = (
            "You specified assume_current but not do_not_scrape_current! "
            "This will result in the same account being collected twice and is likely not what you want."
        )
        core_feedback.error(msg)
        raise ValueError(msg)

    if isinstance(Config.aws.profiles, list) and len(Config.aws.profiles) > 0:
        log.debug("Using specified AWS profiles")
        profiles = Config.aws.profiles
        if Config.aws.account and len(Config.aws.profiles) > 1:
            msg = (
                "You specified both a list of accounts and more than one profile! "
                "This will result in the attempt to collect the same accounts for "
                "every profile and is likely not what you want."
            )
            core_feedback.error(msg)
            raise ValueError(msg)

    for profile in profiles:
        if profile is not None:
            log.debug(f"Finding accounts for profile {profile}")

        try:
            if Config.aws.role and Config.aws.scrape_org:
                log.debug("Role and scrape_org are both set")
                accounts.extend(
                    [
                        AwsAccount(id=aws_account_id, role=Config.aws.role, profile=profile)
                        for aws_account_id in get_org_accounts(
                            filter_current_account=not Config.aws.assume_current,
                            profile=profile,
                            core_feedback=core_feedback,
                        )
                        if aws_account_id not in Config.aws.scrape_exclude_account
                    ]
                )
                if not Config.aws.do_not_scrape_current:
                    accounts.append(AwsAccount(id=current_account_id(profile=profile)))
            elif Config.aws.role and Config.aws.account:
                log.debug("Both, role and list of accounts specified")
                accounts.extend(
                    [
                        AwsAccount(id=aws_account_id, role=Config.aws.role, profile=profile)
                        for aws_account_id in Config.aws.account
                    ]
                )
            else:
                accounts.extend([AwsAccount(id=current_account_id(profile=profile), profile=profile)])
        except botocore.exceptions.NoCredentialsError:
            core_feedback.error(f"No AWS credentials found for {profile}", log)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "AuthFailure":
                core_feedback.error(f"Unable to validate the provided access credentials for {profile}", log)
            elif e.response["Error"]["Code"] == "InvalidClientTokenId":
                core_feedback.error(f"Unable to validate the provided security token for {profile}", log)
            elif e.response["Error"]["Code"] == "ExpiredToken":
                core_feedback.error(f"AWS security token included in the request is expired for {profile}", log)
            elif e.response["Error"]["Code"] == "AccessDenied":
                core_feedback.error(f"Access denied for profile {profile}", log)
            else:
                core_feedback.error(f"AWS client error for profile {profile}: {e}", log)
                raise
        except botocore.exceptions.BotoCoreError as e:
            core_feedback.error(f"Unable to get accounts for profile {profile}: {e}", log)

    set_account_names(accounts)
    return accounts


def get_org_accounts(filter_current_account: bool, profile: Optional[str], core_feedback: CoreFeedback) -> List[str]:
    session = aws_session(profile=profile)
    client = session.client("organizations")
    accounts = []
    try:
        response = client.list_accounts()
        accounts = response.get("Accounts", [])
        while response.get("NextToken") is not None:
            response = client.list_accounts(NextToken=response["NextToken"])
            accounts.extend(response.get("Accounts", []))
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AccessDeniedException":
            core_feedback.error(f"Missing permissions to list organization accounts: {e}", log)
        else:
            raise
    filter_account_id = current_account_id(profile=profile) if filter_current_account else -1
    accounts = [aws_account["Id"] for aws_account in accounts if aws_account["Id"] != filter_account_id]
    for account in accounts:
        log.debug(f"AWS found org account {account}")
    log.info(f"AWS found a total of {len(accounts)} org accounts")
    return accounts


def all_regions(profile: Optional[str] = None) -> List[str]:
    session = aws_session(profile=profile)
    ec2 = session.client("ec2", region_name="us-east-1")
    regions = ec2.describe_regions()
    return [r["RegionName"] for r in regions["Regions"]]


@log_runtime
def collect_account(
    account: AwsAccount,
    regions: List[str],
    args: Namespace,
    running_config: RunningConfig,
    feedback: CoreFeedback,
    cloud: Cloud,
) -> Optional[Graph]:
    collector_name = f"aws_{account.id}"
    resotolib.proc.set_thread_name(collector_name)

    if args is not None:
        ArgumentParser.args = args
        setup_logger("resotoworker-aws", force=True, level=getattr(args, "log_level", None))

    if running_config is not None:
        Config.running_config.apply(running_config)

    if not authenticated(account, feedback):
        log.error(f"Skipping account {account.rtdname}. Reason: authentication failure.")
        return None

    log.debug(f"Starting new collect process for account {account.dname}")

    aac = AwsAccountCollector(Config.aws, cloud, account, regions, feedback)
    try:
        aac.collect()
    except botocore.exceptions.ClientError as e:
        feedback.error(
            f"Ignore account {account.dname}. Reason: An AWS {e.response['Error']['Code']} error occurred.", log
        )
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()
        return None
    except Exception as ex:
        feedback.error(f"Ignore account {account.dname}. Reason: unhandled error occurred: {ex}", log)
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()
        return None

    return aac.graph
